import ast
import copy
import sys
import traceback

from ..java import opcodes as JavaOpcodes
from .modules import Module
from .methods import MainFunction, InitMethod, to_java
from .structures import (
    IF, ELSE, END_IF,
    TRY, CATCH, FINALLY, END_TRY,
    START_LOOP, END_LOOP,
    ArgType,
    jump, OpcodePosition,
    AddToArgs, AddToKwargs
)
from .types.primitives import (
    ASTORE_name, ALOAD_name, free_name,
    ICONST_val, ISTORE_name, ILOAD_name
)
from .types import java, python
from .debug import (
    dump,
    # DEBUG, DEBUG_name
)


def is_call(node, name):
    return (
        # Node is a Call statement...
        isinstance(node, ast.Call)
        # ... where the function being invoked ...
        and isinstance(node.func, ast.Name)
        # ... is the provided name
        and (
            (isinstance(name, str) and node.func.id == name)
            or (isinstance(name, tuple) and node.func.id in name)
        )
    )


def node_visitor(fn):
    def dec(self, node, *args, **kwargs):
        try:
            if node.lineno != self._current_line:
                self._current_line = self.context.next_opcode_starts_line = node.lineno
        except AttributeError:
            pass
        self.context.next_resolve_list.append((node, OpcodePosition.START))
        fn(self, node, *args, **kwargs)
        self.context.next_resolve_list.append((node, OpcodePosition.END))
        self.context.next_resolve_list.append((node, OpcodePosition.NEXT))
    return dec


class NameVisitor(ast.NodeVisitor):
    def evaluate(self, root_node):
        self.names = [[]]
        if root_node:
            self.visit(root_node)
        return self

    @property
    def cls_name(self):
        return '$'.join('.'.join(n for n in group if n) for group in self.names)

    @property
    def ref_name(self):
        return '$'.join('/'.join(n for n in group if n) for group in self.names)

    @property
    def annotation(self):
        type_name = self.ref_name
        return type_name if type_name else 'org/python/Object'

    def visit_Name(self, node):
        self.names[-1].append(node.id)

    def visit_NameConstant(self, node):
        if node.value is None:
            self.names[-1].append("void")
        else:
            raise NotImplementedError("Unknown named constant %s" % node.value)

    def visit_Attribute(self, node):
        self.visit(node.value)
        self.names[-1].append(node.attr)

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.names.append([])
        self.visit(node.slice)

    def visit_Index(self, node):
        self.visit(node.value)


class LocalsVisitor(ast.NodeVisitor):
    def __init__(self, context):
        self.context = context

    def visit_Name(self, node):
        if type(node.ctx) == ast.Store:
            if node.id not in self.context.local_vars:
                self.context.local_vars[node.id] = None

    def visit_Attribute(self, node):
        pass


class Visitor(ast.NodeVisitor):
    def __init__(self, namespace, filename, verbosity=1):
        super().__init__()
        self.namespace = namespace
        self.filename = filename
        self.verbosity = verbosity

        self._current_line = None
        self._context = []
        self._root_module = None

        self.current_exc_name = []

        self.symbol_namespace = {}
        self.code_objects = {}

        self.resolved_yield_expression = {}  # store converted yield expression node identified by generator's id

    @property
    def context(self):
        return self._context[-1]

    @property
    def root_module(self):
        return self._root_module

    def push_context(self, block):
        self._context.append(block)
        block.visitor_setup()

    def pop_context(self):
        self.context.visitor_teardown()
        self._context.pop()

    def full_classref(self, name, default_prefix=None):
        return self.symbol_namespace.get(name, '.'.join([default_prefix, name])).replace('.', '/')

    def extract_code_objects(self, compiled):
        for obj in compiled.co_consts:
            if isinstance(obj, type(compiled)):
                if (obj.co_firstlineno, obj.co_name) in self.code_objects:
                    print(
                        'WARNING: multiple %s code objects found on line %s' % (
                            obj.co_name, obj.co_firstlineno
                        ),
                        file=sys.stderr
                    )
                self.code_objects[(obj.co_firstlineno, obj.co_name)] = obj
                self.extract_code_objects(obj)

    def parse_yield(self, node):
        """parse yield appearing in expression before the expression is evaluated/visited
        """
        def convert_to_Name(_node):
            # convert Yield node to Name node for expression evaluation
            # operand of expression is the message stored in generator
            return ast.Name(
                id='#msg-buffer-%x' % id(_node),
                ctx=ast.Load(),
                lineno=_node.lineno,
                col_offset=_node.col_offset
            )

        for field_name, value in ast.iter_fields(node):
            if isinstance(value, ast.Yield):
                if isinstance(node, ast.Expr):
                    # don't parse Expr(value=Yield)
                    return
                self.visit_Yield(value)
                self.resolved_yield_expression[id(value)] = convert_to_Name(value)
                return
            elif isinstance(value, list):  # args or kwargs list
                index = 0
                for arg_node in value:
                    if isinstance(arg_node, ast.Yield):
                        self.visit_Yield(arg_node)
                        self.resolved_yield_expression[id(arg_node)] = convert_to_Name(arg_node)
                        return
                    index += 1

    def visit(self, node):
        try:
            self.parse_yield(node)
            if id(node) in self.resolved_yield_expression.keys():
                # skip the yield node and visit the converted node (a Name node) instead
                super().visit(self.resolved_yield_expression[id(node)])
                del self.resolved_yield_expression[id(node)]  # delete entry after visited
            else:
                super().visit(node)
        except Exception as e:
            print(
                "Problem occurred in " + str(self.filename)
            )
            print('Node: ' + dump(node))
            raise
        return self._root_module

    def visit_Module(self, node):
        # Compile the module, and extract all the code objects
        compiled = compile(node, filename=self.filename, mode='exec')
        self.extract_code_objects(compiled)

        module = Module(self.namespace, self.filename)
        self.push_context(module)

        if self._root_module is None:
            self._root_module = module

        for child in node.body:
            self.visit(child)

        main = MainFunction(module)
        self.push_context(main)
        # No content, so pop right away.
        # We need to push to make sure setup/teardown
        # logic is invoked.
        self.pop_context()

        module.functions.append(main)

        self.pop_context()

    @node_visitor
    def visit_Interactive(self, node):
        # stmt* body):
        raise NotImplementedError('No handler for Interactive')

    @node_visitor
    def visit_Expr(self, node):
        self.generic_visit(node)

        # If the expression is not yield expression,
        # we need to ignore expression value by popping it off from stack.
        if not isinstance(node.value, ast.Yield):
            self.context.add_opcodes(
                JavaOpcodes.POP()
            )

    @node_visitor
    def visit_Suite(self, node):
        # stmt* body):
        raise NotImplementedError('No handler for Suite')

    # Statements
    @node_visitor
    def visit_FunctionDef(self, node):
        function = self._create_function(node, node.name, node.decorator_list)

        self.push_context(function)

        LocalsVisitor(function).visit(node)

        for child in node.body:
            self.visit(child)

        self.pop_context()

    @node_visitor
    def visit_ClassDef(self, node):
        # Construct a class.
        class_name = node.name

        name_visitor = NameVisitor()

        extends = None
        implements = []

        for keyword in node.keywords:
            key = keyword.arg
            value = keyword.value

            if key == "metaclass":
                raise Exception("Can't handle metaclasses")
            elif key == "extends":
                extends = name_visitor.evaluate(value).ref_name
            elif key == "implements":
                if isinstance(value, ast.List):
                    implements = [
                        name_visitor.visit(v).ref_name
                        for v in value.elts
                    ]
                else:
                    implements = [name_visitor.evaluate(value).ref_name]
            else:
                raise Exception("Unknown meta keyword " + str(key))

        self.context.add_opcodes(
            java.List(),
        )

        if extends:
            first_base = extends
        elif not node.bases:
            first_base = 'org/python/types/Object'
        else:
            first_base = None

        if first_base:
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
                python.Type.for_class(first_base),
                java.List.add(),
            )

        for base in node.bases:
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
            )
            self.visit(base)
            self.context.add_opcodes(
                java.List.add(),
            )

        klass = self.context.add_class(class_name, extends, implements)

        self.push_context(klass)
        for child in node.body:
            self.visit(child)
        self.pop_context()

        self.symbol_namespace[class_name] = klass.class_name

    @node_visitor
    def visit_Return(self, node):
        # expr? value):
        if self.context.generator:
            # PEP 380: return statement in generator is equivalent to raise StopIteration(value)
            if not isinstance(node.value, ast.YieldFrom):
                # Don't close the generator if node.value is ast.YieldFrom,
                # the generator will be closed automatically when node.value is exhausted (raised StopIteration)
                self.context.add_opcodes(
                    ALOAD_name('<generator>'),
                    JavaOpcodes.INVOKEVIRTUAL(
                        'org/python/types/Generator',
                        'close',
                        args=[],
                        returns='Lorg/python/Object;',
                    ),
                    JavaOpcodes.POP(),
                )

            if node.value:
                self.visit(node.value)
            else:
                self.context.add_opcodes(python.NONE())

            self.context.add_opcodes(
                ASTORE_name('#value'),
                java.New('org/python/exceptions/StopIteration'),
                ALOAD_name('#value'),
                java.Init('org/python/exceptions/StopIteration', 'Lorg/python/Object;'),
                JavaOpcodes.ATHROW(),
            )
        elif node.value:
            self.visit(node.value)
            self.context.add_opcodes(JavaOpcodes.ARETURN())
        else:
            self.context.add_opcodes(python.NONE())
            self.context.add_opcodes(JavaOpcodes.ARETURN())

        # Record how deep we were when this return was added.
        self.context.opcodes[-1].needs_implicit_return = \
            self.context.has_nested_structure

    @node_visitor
    def visit_Delete(self, node):
        for target in node.targets:
            self.visit(target)
            if isinstance(target, ast.Attribute):
                self.context.add_opcodes(
                    python.Object.del_attr()
                )
            elif isinstance(target, ast.Subscript):
                self.context.add_opcodes(
                    python.Object.del_item()
                )
            elif isinstance(target, ast.Name):
                # delete is performed by visit(target) with context Del
                pass
            elif isinstance(target, (ast.Tuple, ast.List)):
                self.visit_Delete(ast.Delete(target.elts))
            else:
                raise NotImplementedError('No handler for Delete of type %s' % target)

    @node_visitor
    def visit_Assign(self, node):
        # evaluate the value
        self.visit(node.value)

        if len(node.targets) > 1:
            for target in node.targets:
                # Assign the value to the target
                self.context.add_opcodes(
                    JavaOpcodes.DUP()
                )
                self.visit(target)
            self.context.add_opcodes(
                JavaOpcodes.POP()
            )
        else:
            self.visit(node.targets[0])

    @node_visitor
    def visit_AugAssign(self, node):
        # expr target, operator op, expr value):
        # Evaluate the target
        if isinstance(node.target, ast.Subscript):
            self.visit_Subscript(node.target, ctx=ast.Load())
        elif isinstance(node.target, ast.Attribute):
            self.visit_Attribute(node.target, ctx=ast.Load())
        else:
            self.context.load_name(node.target.id)

        # Evaluate the value
        self.visit(node.value)
        self.context.add_opcodes(
            JavaOpcodes.INVOKEINTERFACE(
                'org/python/Object',
                {
                    ast.Add: '__iadd__',
                    ast.Sub: '__isub__',
                    ast.Mult: '__imul__',
                    ast.Div: '__itruediv__',
                    ast.FloorDiv: '__ifloordiv__',
                    ast.Mod: '__imod__',
                    ast.Pow: '__ipow__',
                    ast.LShift: '__ilshift__',
                    ast.RShift: '__irshift__',
                    ast.BitOr: '__ior__',
                    ast.BitXor: '__ixor__',
                    ast.BitAnd: '__iand__',
                    # ast.MatMult: '__imatmult__',
                }[type(node.op)],
                args=['Lorg/python/Object;'],
                returns='Lorg/python/Object;'
            )
        )
        self.visit(node.target)

    @node_visitor
    def visit_For(self, node):
        self.visit(node.iter)
        self.context.add_opcodes(
            python.Object.iter()
        )

        loop = START_LOOP()

        self.context.add_opcodes(
            JavaOpcodes.ICONST_1(),
            ISTORE_name('#loop-orelse-%x' % id(loop)),
        )
        self.context.store_name('#for-iter-%x' % id(node), declare=True)
        self.context.add_opcodes(
            loop,
        )
        self.context.add_opcodes(
                TRY(),
        )
        self.context.load_name('#for-iter-%x' % id(node)),
        self.context.add_opcodes(
                    python.Iterable.next(),
        )
        self.context.add_opcodes(
                CATCH('org/python/exceptions/StopIteration'),
        )
        self.context.add_opcodes(
                    JavaOpcodes.POP(),
                    jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
                END_TRY(),
        )

        self.visit(node.target)

        for child in node.body:
            self.visit(child)

        self.context.add_opcodes(
            END_LOOP()
        )

        if node.orelse:
            self.context.add_opcodes(
                ILOAD_name('#loop-orelse-%x' % id(loop)),
                IF([], JavaOpcodes.IFEQ)
            )
            for child in node.orelse:
                self.visit(child)
            self.context.add_opcodes(
                END_IF()
            )

        # Clean up
        self.context.delete_name('#for-iter-%x' % id(node))

    @node_visitor
    def visit_While(self, node):
        # expr test, stmt* body, stmt* orelse):

        loop = START_LOOP()

        self.context.add_opcodes(
            JavaOpcodes.ICONST_1(),
            ISTORE_name('#loop-orelse-%x' % id(loop)),
        )

        self.context.add_opcodes(
            loop
        )
        self.visit(node.test)
        self.context.add_opcodes(
            IF([python.Object.as_boolean()], JavaOpcodes.IFNE),
        )
        self.context.add_opcodes(
                jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
            END_IF(),
        )

        for child in node.body:
            self.visit(child)

        self.context.add_opcodes(
            END_LOOP()
        )

        if node.orelse:
            self.context.add_opcodes(
                ILOAD_name('#loop-orelse-%x' % id(loop)),
                IF([], JavaOpcodes.IFEQ)
            )
            for child in node.orelse:
                self.visit(child)
            self.context.add_opcodes(
                END_IF()
            )

    @node_visitor
    def visit_If(self, node):
        # expr test, stmt* body, stmt* orelse):
        self.visit(node.test)

        self.context.add_opcodes(
            IF([python.Object.as_boolean()], JavaOpcodes.IFEQ),
        )

        for child in node.body:
            self.visit(child)

        if node.orelse:
            self.context.add_opcodes(
                ELSE(),
            )

            for child in node.orelse:
                self.visit(child)

        self.context.add_opcodes(
            END_IF()
        )

    @node_visitor
    def visit_With(self, node):
        # withitem* items, stmt* body):
        #     withitem = (expr context_expr, expr? optional_vars)
        exit_names = ['#exit-%d-%x' % (i, id(node)) for i, _ in enumerate(node.items)]

        for i, item in enumerate(node.items):
            self.visit(item.context_expr)

            self.context.add_opcodes(
                JavaOpcodes.DUP(),
                python.Object.get_attribute('__exit__', use_null=True),
                JavaOpcodes.DUP(),
                ASTORE_name(exit_names[i]),
                IF([], JavaOpcodes.IFNONNULL),
                java.THROW(
                    'org/python/exceptions/AttributeError',
                    ['Ljava/lang/String;', JavaOpcodes.LDC_W('__exit__')]
                ),
                END_IF(),
            )

            self.context.add_opcodes(
                python.Object.get_attribute('__enter__', use_null=True),
                JavaOpcodes.DUP(),
                IF([], JavaOpcodes.IFNONNULL),
                java.THROW(
                    'org/python/exceptions/AttributeError',
                    ['Ljava/lang/String;', JavaOpcodes.LDC_W('__enter__')]
                ),
                END_IF(),
            )

            self.context.add_opcodes(
                java.Array(0),
                JavaOpcodes.ACONST_NULL(),
                python.Callable.invoke(),
            )

            if item.optional_vars:
                self.visit(item.optional_vars)
            else:
                self.context.add_opcodes(
                    JavaOpcodes.POP(),
                )

        self.context.add_opcodes(
            TRY(),
        )

        for child in node.body:
            self.visit(child)

        for name in exit_names[::-1]:
            self.context.add_opcodes(
                ALOAD_name(name),
            )
            self.context.add_opcodes(java.Array(3, fill=python.NONE()))
            self.context.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
                python.Callable.invoke(),
                JavaOpcodes.POP(),
            )

        self.context.add_opcodes(
            CATCH('org/python/exceptions/Exception'),
            ASTORE_name('#exception-%x' % id(node)),
        )

        for name in exit_names[::-1]:
            self.context.add_opcodes(
                ALOAD_name(name),
            )
            self.context.add_opcodes(
                java.Array(3),
                JavaOpcodes.DUP(),
                ICONST_val(0),
                python.NONE(),  # TODO: pass the exception type here
                JavaOpcodes.AASTORE(),
                JavaOpcodes.DUP(),
                ICONST_val(1),
                ALOAD_name('#exception-%x' % id(node)),
                JavaOpcodes.AASTORE(),
                JavaOpcodes.DUP(),
                ICONST_val(2),
                python.NONE(),  # TODO: pass the traceback info here
                JavaOpcodes.AASTORE(),
            ),
            self.context.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
                python.Callable.invoke(),
                JavaOpcodes.POP(),
            )

        self.context.add_opcodes(
            ALOAD_name('#exception-%x' % id(node)),
            JavaOpcodes.ATHROW(),
            END_TRY(),
            free_name('#exception-%x' % id(node)),
        )
        for name in exit_names:
            self.context.add_opcodes(free_name(name))

    @node_visitor
    def visit_Raise(self, node):
        if node.exc is None:
            # Re-raise most recent exception.
            self.context.add_opcodes(
                ALOAD_name(self.current_exc_name[-1]),
            )
        else:
            if isinstance(node.exc, ast.Name):
                # handle "raise ValueError" as if "raise ValueError()"
                exception = self.full_classref(node.exc.id, default_prefix='org.python.exceptions')
                self.context.add_opcodes(
                    java.New(exception),
                    java.Init(exception)
                )
            else:
                self.visit(node.exc)

        self.context.add_opcodes(
            JavaOpcodes.CHECKCAST('java/lang/Throwable'),
            JavaOpcodes.ATHROW(),
        )

    @node_visitor
    def visit_Try(self, node):
        # stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody):
        self.context.add_opcodes(
            TRY()
        )

        for child in node.body:
            self.visit(child)

        for child in node.orelse:
            self.visit(child)

        # Finally content is duplicated at the end of the body
        # if it is defined.
        if node.finalbody:
            for child in node.finalbody:
                self.visit(child)

        for handler in node.handlers:
            self.visit(handler)

            # Finally content is duplicated at the end of the handler
            # if it is defined.
            if node.finalbody:
                for child in node.finalbody:
                    self.visit(child)

        if node.finalbody:
            self.context.add_opcodes(
                FINALLY(),
                ASTORE_name('#exception-%x' % id(node))
            )

            for child in node.finalbody:
                self.visit(child)

            self.context.add_opcodes(
                ALOAD_name('#exception-%x' % id(node)),
                JavaOpcodes.ATHROW(),
                free_name('#exception-%x' % id(node))
            )

        self.context.add_opcodes(
            END_TRY()
        )

    @node_visitor
    def visit_Assert(self, node):
        self.visit(node.test)
        self.context.add_opcodes(
            IF([python.Object.as_boolean()], JavaOpcodes.IFNE),
        )
        self.context.add_opcodes(
                java.New('org/python/exceptions/AssertionError'),
        )

        if node.msg:
            self.visit(node.msg)
            self.context.add_opcodes(
                    java.Init('org/python/exceptions/AssertionError', 'Lorg/python/Object;'),
            )
        else:
            self.context.add_opcodes(
                    java.Init('org/python/exceptions/AssertionError'),
            )

        self.context.add_opcodes(
                JavaOpcodes.ATHROW(),
        )
        self.context.add_opcodes(
            END_IF(),
        )

    @node_visitor
    def visit_Import(self, node):
        for alias in node.names:
            self.context.add_opcodes(
                JavaOpcodes.LDC_W(alias.name),
            )
            self.context.load_globals()
            self.context.load_locals()
            self.context.add_opcodes(
                JavaOpcodes.ACONST_NULL(),  # from_list
                JavaOpcodes.ICONST_0(),
                JavaOpcodes.INVOKESTATIC(
                    'org/python/ImportLib',
                    '__import__',
                    args=[
                        'Ljava/lang/String;',
                        'Ljava/util/Map;',
                        'Ljava/util/Map;',
                        '[Ljava/lang/String;',
                        'I',
                    ],
                    returns='Lorg/python/types/Module;'
                )
            )
            if alias.asname:
                self.context.store_name(alias.asname)
            else:
                # The alias will be the fully dotted path. The import
                # will return the top level module. Store the top level
                # module as the top level path.
                self.context.store_name(alias.name.split('.')[0])

    @node_visitor
    def visit_ImportFrom(self, node):
        if node.module:
            self.context.add_opcodes(
                JavaOpcodes.LDC_W(node.module),
            )
        else:
            self.context.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
            )

        self.context.load_globals()
        self.context.load_locals()

        self.context.add_opcodes(
            java.Array(len(node.names), 'java/lang/String'),
        )

        for i, alias in enumerate(node.names):
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
                ICONST_val(i),
                JavaOpcodes.LDC_W(alias.name),
                JavaOpcodes.AASTORE(),
            )

        self.context.add_opcodes(
            ICONST_val(node.level),
            JavaOpcodes.INVOKESTATIC(
                'org/python/ImportLib',
                '__import__',
                args=[
                    'Ljava/lang/String;',
                    'Ljava/util/Map;',
                    'Ljava/util/Map;',
                    '[Ljava/lang/String;',
                    'I',
                ],
                returns='Lorg/python/types/Module;'
            )
        )

        if len(node.names) == 1 and node.names[0].name == '*':
            # Find exported symbols (__all__, or everything but private "_" symbols)
            self.context.add_opcodes(
                JavaOpcodes.INVOKESTATIC(
                    'org/python/ImportLib',
                    'importAll',
                    args=['Lorg/python/types/Module;'],
                    returns='Ljava/util/Map;'
                )
            )

            # Add all the exported symbols to the currrent context
            self.context.store_dynamic()
        else:
            for alias in node.names:
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    python.Object.get_attribute(alias.name),
                )

                if alias.asname:
                    self.context.store_name(alias.asname)
                else:
                    self.context.store_name(alias.name)

            self.context.add_opcodes(
                JavaOpcodes.POP(),
            )

    @node_visitor
    def visit_Global(self, node):
        for name in node.names:
            self.context.local_vars.pop(name, None)

    @node_visitor
    def visit_Nonlocal(self, node):
        # identifier* names):
        for name in node.names:
            self.context.nonlocal_vars.append(name)
            self.context.local_vars.pop(name, None)

    @node_visitor
    def visit_Pass(self, node):
        pass

    @node_visitor
    def visit_Break(self, node):
        for loop in self.context.loops[::-1]:
            if loop.end_op is None:
                break

        self.context.add_opcodes(
            JavaOpcodes.ICONST_0(),
            ISTORE_name('#loop-orelse-%x' % id(loop)),
        )
        self.context.add_opcodes(
            jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )

    @node_visitor
    def visit_Continue(self, node):
        for loop in self.context.loops[::-1]:
            if loop.end_op is None:
                break
        self.context.add_opcodes(
            jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.START),
        )

    # # Expressions
    @node_visitor
    def visit_BoolOp(self, node):
        # boolop op, expr* values):
        comparison = {
            ast.Or: JavaOpcodes.IFNE,
            ast.And: JavaOpcodes.IFEQ,
        }[type(node.op)]

        for child in node.values[:-1]:
            self.visit(child)

            self.context.add_opcodes(
                # Duplicate the value found by the child, and evaluate
                # it's truthiness. If it matches the boolean operation,
                # we've found a match; jump to the end.
                JavaOpcodes.DUP(),
                python.Object.as_boolean(),
                jump(comparison(0), self.context, node, OpcodePosition.END),

                # This value wasn't a match; pop it off the stack.
                JavaOpcodes.POP(),
            )
        # The last value is the failsafe.
        self.visit(node.values[-1])

    @node_visitor
    def visit_BinOp(self, node):
        # expr left, operator op, expr right):
        if isinstance(node.op, ast.Pow):
            self.visit(node.left)
            self.visit(node.right)
            self.context.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__pow__',
                    args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                    returns='Lorg/python/Object;'
                ),
            )
        else:
            self.visit(node.left)
            self.visit(node.right)
            self.context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    {
                        ast.Add: '__add__',
                        ast.Sub: '__sub__',
                        ast.Mult: '__mul__',
                        ast.Div: '__truediv__',
                        ast.FloorDiv: '__floordiv__',
                        ast.Mod: '__mod__',
                        ast.Pow: '__pow__',
                        ast.LShift: '__lshift__',
                        ast.RShift: '__rshift__',
                        ast.BitOr: '__or__',
                        ast.BitXor: '__xor__',
                        ast.BitAnd: '__and__',
                        # ast.MatMult:
                    }[type(node.op)],
                    args=['Lorg/python/Object;'],
                    returns='Lorg/python/Object;'
                )
            )

    @node_visitor
    def visit_UnaryOp(self, node):
        self.visit(node.operand)
        self.context.add_opcodes(
            JavaOpcodes.INVOKEINTERFACE(
                'org/python/Object',
                {
                    ast.USub: '__neg__',
                    ast.UAdd: '__pos__',
                    ast.Not: '__not__',
                    ast.Invert: '__invert__',
                }[type(node.op)],
                args=[],
                returns='Lorg/python/Object;'
            )
        )

    def _create_function(self, node, func_name, decorator_list):
        name_visitor = NameVisitor()
        default_vars = []
        parameter_signatures = []
        arg_index = {}
        for i, arg in enumerate(node.args.args):
            index = len(node.args.defaults) - len(node.args.args) + i
            if index != 0:
                # Store the index of the arguments; this is only used by
                # the @super decorator on the constructors.
                # self can be ignored.
                arg_index[arg.arg] = i - 1
            if index >= 0:
                default = '#%s-default-%s-%x' % (func_name, i, id(node))
                self.visit(node.args.defaults[index])
                self.context.add_opcodes(
                    ASTORE_name(default)
                )
                default_vars.append(default)
            else:
                default = None

            parameter_signatures.append({
                'name': arg.arg,
                'annotation': name_visitor.evaluate(arg.annotation).annotation,
                'kind': ArgType.POSITIONAL_OR_KEYWORD,
                'default': default
            })

        if node.args.vararg:
            arg_index[node.args.vararg] = len(node.args.args)
            parameter_signatures.append({
                'name': node.args.vararg.arg,
                'annotation': name_visitor.evaluate(node.args.vararg.annotation).annotation,
                'kind': ArgType.VAR_POSITIONAL,
            })

        for i, arg in enumerate(node.args.kwonlyargs):
            index = len(node.args.kw_defaults) - len(node.args.kwonlyargs) + i
            arg_index[arg.arg] = None
            if index >= 0 and node.args.kw_defaults[index] is not None:
                default = '#%s-kw_default-%s-%x' % (func_name, i, id(node))
                self.visit(node.args.kw_defaults[index])
                self.context.add_opcodes(
                    ASTORE_name(default)
                )
                default_vars.append(default)
            else:
                default = None

            parameter_signatures.append({
                'name': arg.arg,
                'annotation': name_visitor.evaluate(arg.annotation).annotation,
                'kind': ArgType.KEYWORD_ONLY,
                'default': default
            })

        if node.args.kwarg:
            arg_index[node.args.kwarg] = None
            parameter_signatures.append({
                'name': node.args.kwarg.arg,
                'annotation': name_visitor.evaluate(node.args.kwarg.annotation).annotation,
                'kind': ArgType.VAR_KEYWORD,
            })

        returns = getattr(node, 'returns', None)
        return_signature = {
            'annotation': name_visitor.evaluate(returns).annotation
        }

        # Now actually define the function.
        for decorator in decorator_list:
            # The @super decorator on __init__() is a special case. It tells us how
            # we want to invoke super() in the Java construction process.
            # It should have a single argument; a dictionary that has the
            # expressions to be passed in as arguments as keys, and the type
            # annotation for the argument as the value. For example:
            #
            #   class MyClass(BaseClass):
            #       @super({value: int, value*2: float, "Hello": java.lang.String})
            #       def __init__(self, value):
            #           ...
            #
            # would map to the rough equivalent of:
            #
            #   public MyClass(value) {
            #       super(value, value * 2, "Hello");
            #           ...
            #   }
            #
            if is_call(decorator, 'super'):
                if func_name == '__init__':
                    if len(decorator.args) == 1 and isinstance(decorator.args[0], ast.Dict):
                        super_args = [
                            name_visitor.evaluate(arg_type).ref_name
                            for arg_type in decorator.args[0].values
                        ]
                        init = InitMethod(
                            klass=self.context,
                            args=arg_index,
                            super_args=super_args,
                        )
                        self.push_context(init)

                        for arg, annotation in zip(decorator.args[0].keys, super_args):
                            self.visit(arg)
                            to_java(self.context, annotation)

                        self.pop_context()
                        self.context.constructor = init
                    else:
                        raise Exception("super() (as __init__ decorator) expects a single dictionary as an argument")
                else:
                    raise Exception("super() can only be used as a decorator on __init__()")
            elif isinstance(decorator, ast.Name) and decorator.id == "classmethod":
                # print("FIXME: Ignoring classmethod")
                pass
            elif isinstance(decorator, ast.Name) and decorator.id == "staticmethod":
                # print("FIXME: Ignoring staticmethod")
                pass
            else:
                self.visit(decorator)
                self.context.add_opcodes(
                    JavaOpcodes.CHECKCAST('org/python/Callable'),
                    java.Array(1),
                    JavaOpcodes.DUP(),
                    JavaOpcodes.ICONST_0(),
                )

        function = self.context.add_function(
            name=func_name,
            code=self.code_objects[(node.lineno, getattr(node, 'name', '<lambda>'))],
            parameter_signatures=parameter_signatures,
            return_signature=return_signature,
        )

        for decorator in decorator_list[::-1]:
            if func_name == '__init__' and is_call(decorator, 'super'):
                # We can ignore the @super decorator on __init__ methods.
                pass
            elif isinstance(decorator, ast.Name) and decorator.id == "classmethod":
                pass
            elif isinstance(decorator, ast.Name) and decorator.id == "staticmethod":
                pass
            else:
                self.context.add_opcodes(
                    JavaOpcodes.AASTORE(),
                    JavaOpcodes.ACONST_NULL(),
                    python.Callable.invoke(),
                )

        # Store the callable object as an accessible symbol.
        self.context.store_name(func_name)

        # Free all the variables used for default storage.
        for default in default_vars:
            self.context.add_opcodes(
                free_name(default)
            )

        return function

    @node_visitor
    def visit_Lambda(self, node):
        lambda_name = 'lambda-%x' % id(node)
        function = self._create_function(node, lambda_name, [])

        self.push_context(function)

        LocalsVisitor(function).visit(node)

        self.visit(node.body)
        self.context.add_opcodes(JavaOpcodes.ARETURN())
        self.context.opcodes[-1].needs_implicit_return = \
            self.context.has_nested_structure

        self.pop_context()

        self.context.load_name(lambda_name)

    @node_visitor
    def visit_IfExp(self, node):
        self.visit(node.test)

        self.context.add_opcodes(
            IF([python.Object.as_boolean()], JavaOpcodes.IFEQ),
        )

        self.visit(node.body)

        self.context.add_opcodes(
            ELSE(),
        )

        self.visit(node.orelse)

        self.context.add_opcodes(
            END_IF(),
        )

    @node_visitor
    def visit_Dict(self, node):
        self.context.add_opcodes(
            python.Dict()
        )

        for kchild, vchild in zip(node.keys, node.values):
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
            )

            self.visit(kchild)
            self.visit(vchild)

            self.context.add_opcodes(
                python.Dict.set_item(),
            )

        self.context.add_opcodes(
        )

    @node_visitor
    def visit_Set(self, node):
        self.context.add_opcodes(
            python.Set()
        )

        for child in node.elts:
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
            )

            self.visit(child)

            self.context.add_opcodes(
                python.Set.add()
            )

    @node_visitor
    def visit_ListComp(self, node):
        listcomp_name = 'listcomp_%x' % id(node)
        listcomp = self.context.add_function(
            name=listcomp_name,
            code=self.code_objects[(node.lineno, '<listcomp>')],
            parameter_signatures=[
                {
                    'name': '.%s' % i,
                    # 'annotation': name_visitor.evaluate(arg.annotation).annotation,
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'default': None
                }
                for i, arg in enumerate(node.generators)
            ],
            return_signature={
                'annotation': 'org/python/types/List'
            }
        )

        # Store the callable object as an accessible symbol.
        self.context.store_name(listcomp_name)

        self.push_context(listcomp)

        LocalsVisitor(listcomp).visit(node)

        self.context.add_opcodes(
            python.List(),
        )
        self.context.store_name('#listcomp-result-%x' % id(node), declare=True)

        if len(node.generators) != 1:
            raise NotImplementedError("Don't know how to handle multiple generators")

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    ALOAD_name('.%s' % i),
                    python.Object.iter(),
                )
            else:
                raise NotImplementedError("Don't know how to handle generator of type %s" % type(generator))

        loop = START_LOOP()

        self.context.store_name('#listcomp-iter-%x' % id(node), declare=True)
        self.context.add_opcodes(
            loop,
        )
        self.context.add_opcodes(
                TRY(),
        )
        self.context.load_name('#listcomp-iter-%x' % id(node))
        self.context.add_opcodes(
                    python.Iterable.next(),
        )
        self.context.add_opcodes(
                CATCH('org/python/exceptions/StopIteration'),
        )
        self.context.add_opcodes(
                    JavaOpcodes.POP(),
                    jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
                END_TRY(),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.visit(generator.target)

        if node.generators[0].ifs:
            self.visit(
                ast.BoolOp(ast.And(), node.generators[0].ifs)
            )

            self.context.add_opcodes(
                IF([python.Object.as_boolean()], JavaOpcodes.IFEQ),
            )

        self.visit(node.elt)

        # And add it to the result list
        self.context.load_name('#listcomp-result-%x' % id(node)),
        self.context.add_opcodes(
                JavaOpcodes.CHECKCAST('org/python/types/List'),
                JavaOpcodes.SWAP(),
                python.List.append(),
        )

        if node.generators[0].ifs:
            self.context.add_opcodes(
                END_IF()
            )

        self.context.add_opcodes(
            END_LOOP(),
        )
        self.context.load_name('#listcomp-result-%x' % id(node)),
        self.context.add_opcodes(
            JavaOpcodes.ARETURN(),
        )

        # Clean up
        self.context.delete_name('#listcomp-iter-%x' % id(node))
        self.context.delete_name('#listcomp-result-%x' % id(node))

        self.pop_context()

        # Now invoke the list comprehension
        self.context.load_name('listcomp_%x' % id(node))
        self.context.add_opcodes(
            java.Array(len(node.generators))
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    ICONST_val(i),
                )

                self.visit(generator.iter)

                self.context.add_opcodes(
                    JavaOpcodes.AASTORE(),
                )

        self.context.add_opcodes(
            # No keyword arguments
            JavaOpcodes.ACONST_NULL(),

            # Now invoke.
            python.Callable.invoke(),
        )

        # FIXME: This would be a much more efficient way to invoke
        # the comprehension method, but when it's a closure, it's
        # not that simple.
        # for i, generator in enumerate(node.generators):
        #     if isinstance(generator, ast.comprehension):
        #         self.visit(generator.iter)
        #
        # self.context.add_opcodes(
        #     JavaOpcodes.INVOKESTATIC(
        #         self.context.class_descriptor,
        #         'listcomp_%x' % id(node),
        #         '(%s)Lorg/python/Object;' % ''.join(
        #             'Lorg/python/Object;'
        #             for gen in node.generators
        #         )
        #     )
        # )

    @node_visitor
    def visit_SetComp(self, node):
        # Get the code object for the list comprehension.
        setcomp_name = 'setcomp_%x' % id(node)
        setcomp = self.context.add_function(
            name=setcomp_name,
            code=self.code_objects[(node.lineno, '<setcomp>')],
            parameter_signatures=[
                {
                    'name': '.%s' % i,
                    # 'annotation': name_visitor.evaluate(arg.annotation).annotation,
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'default': None
                }
                for i, arg in enumerate(node.generators)
            ],
            return_signature={
                'annotation': 'org/python/types/Set'
            }
        )
        # Store the callable object as an accessible symbol.
        self.context.store_name(setcomp_name)

        self.push_context(setcomp)

        LocalsVisitor(setcomp).visit(node)

        self.context.add_opcodes(
            python.Set(),
        )
        self.context.store_name('#setcomp-result-%x' % id(node), declare=True)

        if len(node.generators) != 1:
            raise NotImplementedError("Don't know how to handle multiple generators")

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    ALOAD_name('.%s' % i),
                    python.Object.iter()
                )
            else:
                raise NotImplementedError("Don't know how to handle generator of type %s" % type(generator))

        self.context.store_name('#setcomp-iter-%x' % id(node), declare=True)

        loop = START_LOOP()
        self.context.add_opcodes(
            loop,
        )
        self.context.add_opcodes(
                TRY(),
        )
        self.context.load_name('#setcomp-iter-%x' % id(node))
        self.context.add_opcodes(
                    python.Iterable.next(),
        )
        self.context.add_opcodes(
                CATCH('org/python/exceptions/StopIteration'),
        )
        self.context.add_opcodes(
                    JavaOpcodes.POP(),
                    jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
                END_TRY(),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.visit(generator.target)

        self.visit(node.elt)

        # And add it to the result set
        self.context.load_name('#setcomp-result-%x' % id(node))
        self.context.add_opcodes(
                JavaOpcodes.CHECKCAST('org/python/types/Set'),
                JavaOpcodes.SWAP(),
                python.Set.add(),
        )
        self.context.add_opcodes(
            END_LOOP(),
        )
        self.context.load_name('#setcomp-result-%x' % id(node)),
        self.context.add_opcodes(
            JavaOpcodes.ARETURN(),
        )

        # Clean up
        self.context.delete_name('#setcomp-iter-%x' % id(node))
        self.context.delete_name('#setcomp-result-%x' % id(node))

        self.pop_context()

        # Now invoke the set comprehension
        self.context.load_name('setcomp_%x' % id(node))
        self.context.add_opcodes(
            java.Array(len(node.generators)),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    ICONST_val(i),
                )

                self.visit(generator.iter)

                self.context.add_opcodes(
                    JavaOpcodes.AASTORE(),
                )

        self.context.add_opcodes(
            # No keyword arguments
            JavaOpcodes.ACONST_NULL(),

            # Now invoke.
            python.Callable.invoke(),
        )

    @node_visitor
    def visit_DictComp(self, node):
        dictcomp_name = 'dictcomp_%x' % id(node)
        dictcomp = self.context.add_function(
            name=dictcomp_name,
            code=self.code_objects[(node.lineno, '<dictcomp>')],
            parameter_signatures=[
                {
                    'name': '.%s' % i,
                    # 'annotation': name_visitor.evaluate(arg.annotation).annotation,
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'default': None
                }
                for i, arg in enumerate(node.generators)
            ],
            return_signature={
                'annotation': 'org/python/types/Dict'
            }
        )
        # Store the callable object as an accessible symbol.
        self.context.store_name(dictcomp_name)

        self.push_context(dictcomp)

        LocalsVisitor(dictcomp).visit(node)

        self.context.add_opcodes(
            java.New('org/python/types/Dict'),
            java.Init('org/python/types/Dict'),
        )
        self.context.store_name('#dictcomp-result-%x' % id(node), declare=True)

        if len(node.generators) != 1:
            raise NotImplementedError("Don't know how to handle multiple generators")

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    ALOAD_name('.%s' % i),
                    python.Object.iter()
                )
            else:
                raise NotImplementedError("Don't know how to handle generator of type %s" % type(generator))

        self.context.store_name('#dictcomp-iter-%x' % id(node), declare=True)

        loop = START_LOOP()
        self.context.add_opcodes(
            loop,
        )
        self.context.add_opcodes(
                TRY(),
        )
        self.context.load_name('#dictcomp-iter-%x' % id(node)),
        self.context.add_opcodes(
                    python.Iterable.next(),
        )
        self.context.add_opcodes(
                CATCH('org/python/exceptions/StopIteration'),
        )
        self.context.add_opcodes(
                    JavaOpcodes.POP(),
                    jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
                END_TRY(),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.visit(generator.target)

        self.visit(node.key)
        self.context.add_opcodes(
            ASTORE_name('#dictcomp-key-%x' % id(node)),
        )

        self.visit(node.value)
        self.context.add_opcodes(
            ASTORE_name('#dictcomp-value-%x' % id(node)),
        )

        # And add it to the result list
        self.context.load_name('#dictcomp-result-%x' % id(node))
        self.context.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Dict'),
            ALOAD_name('#dictcomp-key-%x' % id(node)),
            ALOAD_name('#dictcomp-value-%x' % id(node)),
            python.Dict.set_item(),
            END_LOOP(),
        )
        self.context.load_name('#dictcomp-result-%x' % id(node))
        self.context.add_opcodes(
            JavaOpcodes.ARETURN(),

            # Clean up
            free_name('#dictcomp-key-%x' % id(node)),
            free_name('#dictcomp-value-%x' % id(node))
        )

        self.context.delete_name('#dictcomp-iter-%x' % id(node))
        self.context.delete_name('#dictcomp-result-%x' % id(node))

        self.pop_context()

        # Now invoke the dict comprehension
        self.context.load_name('dictcomp_%x' % id(node))
        self.context.add_opcodes(
            java.Array(len(node.generators)),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    ICONST_val(i),
                )

                self.visit(generator.iter)

                self.context.add_opcodes(
                    JavaOpcodes.AASTORE(),
                )

        self.context.add_opcodes(
            # No keyword arguments
            JavaOpcodes.ACONST_NULL(),

            # Now invoke.
            python.Callable.invoke(),
        )

    @node_visitor
    def visit_GeneratorExp(self, node):
        genexp_name = 'genexp_%x' % id(node)
        genexp = self.context.add_function(
            name=genexp_name,
            code=self.code_objects[(node.lineno, '<genexpr>')],
            parameter_signatures=[
                {
                    'name': '.%s' % i,
                    # 'annotation': name_visitor.evaluate(arg.annotation).annotation,
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'default': None
                }
                for i, arg in enumerate(node.generators)
            ],
            return_signature={
                'annotation': 'org/python/types/List'
            }
        )

        # Store the callable object as an accessible symbol.
        self.context.store_name(genexp_name)

        self.push_context(genexp)

        LocalsVisitor(genexp).visit(node)

        if len(node.generators) != 1:
            raise NotImplementedError("Don't know how to handle multiple generators")

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    ALOAD_name('.%s' % i),
                    python.Object.iter()
                )
            else:
                raise NotImplementedError("Don't know how to handle generator of type %s" % type(generator))

        self.context.store_name('#genexp-iter-%x' % id(node), declare=True)
        loop = START_LOOP()
        self.context.add_opcodes(
            loop,
        )
        self.context.add_opcodes(
                TRY(),
        )
        self.context.load_name('#genexp-iter-%x' % id(node))
        self.context.add_opcodes(
                    python.Iterable.next(),
        )
        self.context.add_opcodes(
                CATCH('org/python/exceptions/StopIteration'),
        )
        self.context.add_opcodes(
                    JavaOpcodes.POP(),
                    jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),
        )
        self.context.add_opcodes(
                END_TRY(),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.visit(generator.target)

        self.visit(node.elt)

        yield_point = len(self.context.yield_points) + 1
        self.context.add_opcodes(
            # Save the current stack and yield index
            ALOAD_name('<generator>'),
            ALOAD_name('#locals'),
            java.Yield(yield_point),
        )

        # On restore, the next instruction is the target
        # for the restore jump.
        self.context.yield_points.append(node)
        self.context.next_resolve_list.append((node, OpcodePosition.YIELD))

        #  First thing to do is restore the state of the stack.
        self.context.add_opcodes(
            ALOAD_name('<generator>'),
            JavaOpcodes.GETFIELD('org/python/types/Generator', 'stack', 'Ljava/util/Map;'),
            ASTORE_name('#locals'),
        )

        for var, index in self.context.local_vars.items():
            if index is not None and var not in ('<generator>', '#locals'):
                self.context.add_opcodes(
                    ALOAD_name('#locals'),
                    java.Map.get(var),
                    JavaOpcodes.ASTORE(index),
                )

        self.context.add_opcodes(
            END_LOOP(),
        )

        # Clean up
        self.context.delete_name('#genexp-iter-%x' % id(node))

        self.pop_context()

        # Now invoke the list comprehension
        self.context.load_name('genexp_%x' % id(node))
        self.context.add_opcodes(
            java.Array(len(node.generators)),
        )

        for i, generator in enumerate(node.generators):
            if isinstance(generator, ast.comprehension):
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    ICONST_val(i),
                )

                self.visit(generator.iter)

                self.context.add_opcodes(
                    JavaOpcodes.AASTORE(),
                )

        self.context.add_opcodes(
            # No keyword arguments
            JavaOpcodes.ACONST_NULL(),

            # Now invoke.
            python.Callable.invoke(),
        )

    @node_visitor
    def visit_Yield(self, node):
        if hasattr(node, "lineno"):
            # Checks for programmatically defined yield node.
            # A programmatically defined yield node does not has the attribute `lineno`
            # Example: currently only visit_YieldFrom function define yield node programmatically
            # (defined as ast.Yield(None)) before pushing next yield value on stack
            if node.value:
                self.visit(node.value)
            else:
                # push NoneType object to stack to support single yield statement (without operand)
                # as the statment `yield` is equivalent to `yield None`
                self.context.add_opcodes(
                    JavaOpcodes.GETSTATIC(
                        'org/python/types/NoneType',
                        'NONE',
                        'Lorg/python/Object;'
                    )
                )
                
        yield_point = len(self.context.yield_points) + 1

        # Save the current stack and yield index
        self.context.load_name('<generator>')
        self.context.add_opcodes(
            ALOAD_name('#locals'),
            java.Yield(yield_point),
        )

        # On restore, the next instruction is the target
        # for the restore jump.
        node_to_resolve = copy.deepcopy(node)  # deepcopy is required for multiple visits of finalbody from visitTry
        self.context.yield_points.append(node_to_resolve)
        self.context.next_resolve_list.append((node_to_resolve, OpcodePosition.YIELD))

        #  First thing to do is restore the state of the stack.
        self.context.load_name('<generator>')
        self.context.add_opcodes(
            JavaOpcodes.GETFIELD('org/python/types/Generator', 'stack', 'Ljava/util/Map;'),
            ASTORE_name('#locals'),
        )

        for var, index in self.context.local_vars.items():
            if index is not None and var not in ('<generator>', '#locals') \
                    and "#exception-" not in var:  # Don't load exception
                self.context.add_opcodes(
                    ALOAD_name('#locals'),
                    java.Map.get(var),
                    JavaOpcodes.ASTORE(index),
                )

        # Manage messages and exception when generator is restored
        if hasattr(node, "lineno"):
            # Retrieve message from generator
            self.context.load_name('<generator>')
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
                JavaOpcodes.DUP(),
                JavaOpcodes.GETFIELD('org/python/types/Generator', 'message', 'Lorg/python/Object;'),
            )
            # Store the message in local variable for expression evaluation
            self.context.store_name('#msg-buffer-%x' % id(node))
            self.context.add_opcodes(
                # Reset message to None after retrieval
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Generator',
                    'reset_message',
                    args=[],
                    returns='V',
                ),
                # Throw exception if there is one
                # NO-OP if generator.exception is null
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Generator',
                    'throw_exception',
                    args=[],
                    returns='V'
                )
            )

    @node_visitor
    def visit_YieldFrom(self, node):
        self.visit(node.value)  # pushes expression on stack
        self.context.add_opcodes(
            python.Object.iter(),  # pops expression from stack then gets and pushes its iterator on stack
        )

        # invokes __next()__ on the iterator, return if StopIteration
        self.context.store_name('#yield-iter-%x' % id(node))
        self.context.load_name('#yield-iter-%x' % id(node))
        self.context.add_opcodes(
            TRY(),
            python.Iterable.next(),
            ASTORE_name('#yield-value-%x' % id(node)),
            CATCH('org/python/exceptions/StopIteration'),
            JavaOpcodes.GETFIELD('org/python/exceptions/StopIteration', 'value', 'Lorg/python/Object;'),
            JavaOpcodes.ARETURN(),
            END_TRY()
        )

        # Yield the value and intercepts exception by invoking generator.intercept_exception
        # If the method throws StopIteration, get its value and break out of the loop
        # If the method returns org.python.Object, store the value for next yield and continue loop
        loop = START_LOOP()
        self.context.add_opcodes(
            loop,
            ALOAD_name('#yield-value-%x' % id(node))
        )
        self.visit_Yield(ast.Yield(None))
        self.context.load_name('<generator>')
        self.context.load_name('#yield-iter-%x' % id(node))
        self.context.add_opcodes(
            TRY(),
            JavaOpcodes.INVOKEVIRTUAL(
                'org/python/types/Generator',
                'intercept_exception',
                args=['Lorg/python/Object;'],
                returns='Lorg/python/Object;'
            ),
            CATCH('org/python/exceptions/StopIteration'),
            python.Object.get_attribute('value'),
            jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.NEXT),  # break from the loop
            END_TRY(),

            ASTORE_name('#yield-value-%x' % id(node)),
            jump(JavaOpcodes.GOTO(0), self.context, loop, OpcodePosition.START),  # continue
            END_LOOP(),
        )

    @node_visitor
    def visit_Compare(self, node):

        def compare_to(arg):
            if isinstance(arg, ast.Is) and not const_comparison:
                self.context.add_opcodes(
                    IF([], JavaOpcodes.IF_ACMPNE),
                )
                self.context.add_opcodes(
                        JavaOpcodes.GETSTATIC('org/python/types/Bool', 'TRUE', 'Lorg/python/types/Bool;'),
                )
                self.context.add_opcodes(
                    ELSE(),
                )
                self.context.add_opcodes(
                        JavaOpcodes.GETSTATIC('org/python/types/Bool', 'FALSE', 'Lorg/python/types/Bool;'),

                )
                self.context.add_opcodes(
                    END_IF(),
                )

            elif isinstance(arg, ast.IsNot) and not const_comparison:
                self.context.add_opcodes(
                    IF([], JavaOpcodes.IF_ACMPEQ),
                )
                self.context.add_opcodes(
                        JavaOpcodes.GETSTATIC('org/python/types/Bool', 'TRUE', 'Lorg/python/types/Bool;'),
                )
                self.context.add_opcodes(
                    ELSE(),
                )
                self.context.add_opcodes(
                        JavaOpcodes.GETSTATIC('org/python/types/Bool', 'FALSE', 'Lorg/python/types/Bool;'),
                )
                self.context.add_opcodes(
                    END_IF(),
                )

            else:
                if isinstance(arg, (ast.Eq, ast.Is)):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__eq__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, (ast.NotEq, ast.IsNot)):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__ne__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.Lt):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__lt__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.LtE):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__le__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.Gt):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__gt__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.GtE):
                    self.context.add_opcodes(
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__ge__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.In):
                    self.context.add_opcodes(
                        JavaOpcodes.SWAP(),
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__contains__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

                if isinstance(arg, ast.NotIn):
                    self.context.add_opcodes(
                        JavaOpcodes.SWAP(),
                        JavaOpcodes.INVOKESTATIC(
                            'org/python/types/Object',
                            '__not_contains__',
                            args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                            returns='Lorg/python/Object;'),
                    )

        self.visit(node.left)
        left = node.left

        for i, (operation, right) in enumerate(zip(node.ops, node.comparators)):
            self.visit(right)
            const_comparison = isinstance(left, ast.Num) | isinstance(right, ast.Num)
            if i < len(node.ops) - 1:
                self.context.add_opcodes(
                        JavaOpcodes.DUP_X1()
                        )
                compare_to(operation)
                self.context.add_opcodes(
                    JavaOpcodes.DUP()
                )
                self.context.add_opcodes(
                    IF([python.Object.as_boolean()], JavaOpcodes.IFNE)
                )
                self.context.add_opcodes(
                        JavaOpcodes.SWAP(),
                        JavaOpcodes.POP()
                )
                self.context.add_opcodes(
                    ELSE()
                )
                self.context.add_opcodes(
                        JavaOpcodes.POP()
                )
            else:
                compare_to(operation)
            left = right

        for _ in range(len(node.ops) - 1):
            self.context.add_opcodes(
                END_IF()
            )

    @node_visitor
    def visit_Call(self, node):

        if is_call(node, ('locals', 'globals', 'vars')):
            # **kwargs is node.keywords[None] in Python 3.5; node.kwargs in earlier versions
            if None in node.keywords or getattr(node, 'kwargs', None):
                self.context.add_opcodes(
                    java.New('org/python/exceptions/TypeError'),
                    JavaOpcodes.LDC_W(node.func.id + "() takes no keyword arguments"),
                    java.Init('org/python/exceptions/TypeError', 'Ljava/lang/String;'),
                    JavaOpcodes.ATHROW()
                )
            elif node.args:
                self.context.add_opcodes(
                    java.New('org/python/exceptions/NotImplementedError'),
                    JavaOpcodes.LDC_W(node.func.id + "(object) not yet implemented"),
                    java.Init('org/python/exceptions/NotImplementedError', 'Ljava/lang/String;'),
                    JavaOpcodes.ATHROW()
                )
            else:
                # Create a dict for storage
                self.context.add_opcodes(
                    java.New('org/python/types/Dict'),
                    java.New('org/python/internals/Scope'),
                )

                getattr(self.context, 'load_%s' % node.func.id)()

                self.context.add_opcodes(
                    # Wrap the locals/globals/vars to make them look like a Python String->Object map
                    java.Init('org/python/internals/Scope', 'Ljava/util/Map;'),
                    # Construct a dictionary based on that map
                    java.Init('org/python/types/Dict', 'Ljava/util/Map;'),
                )

        elif is_call(node, 'super'):
            # context.add_opcodes(
            #     DEBUG("ATTRIBUTE ON SUPER"),
            # )

            if len(node.args) == 0:
                self.context.add_opcodes(
                    java.New('org/python/types/Super'),

                    # The super class to bind to.
                    python.Type.for_name(self.context.klass.descriptor),

                    # Bind to self. Since we know we are in a class building context,
                    # we can be certain that register 0 contains self.
                    JavaOpcodes.ALOAD_0(),
                    java.Init('org/python/types/Super', 'Lorg/python/Object;', 'Lorg/python/Object;'),
                )

            elif len(node.args) == 1:  # Unbound super
                # FIXME: This is an unusual call pattern for super(), so it
                # isn't obvious what the right behavior should be...
                self.context.add_opcodes(
                    java.New('org/python/types/Super'),
                )

                # The super class to bind to.
                self.visit(node.args[0])

                self.context.add_opcodes(
                    java.Init('org/python/types/Super', 'Lorg/python/Object;'),
                )

            elif len(node.args) == 2:  # Bound super
                self.context.add_opcodes(
                    java.New('org/python/types/Super'),
                )

                # The super class to bind to.
                self.visit(node.args[0])

                # The instance to bind to.
                self.visit(node.args[1])

                self.context.add_opcodes(
                    java.Init('org/python/types/Super', 'Lorg/python/Object;', 'Lorg/python/Object;'),
                )

            else:
                raise Exception("Invalid number of arguments to super()")

        else:
            # Evaluate the callable, and check that it *is* a callable
            self.visit(node.func)

            self.context.add_opcodes(
                JavaOpcodes.CHECKCAST('org/python/Callable'),
            )

            # Create and populate the array of arguments to pass to invoke()
            num_args = len([arg for arg in node.args if not isinstance(arg, ast.Starred)])

            if len(node.args) == 0 and getattr(node, 'starargs', None) is None:
                self.context.add_opcodes(
                    JavaOpcodes.ACONST_NULL(),
                )
            else:
                self.context.add_opcodes(
                    java.Array(num_args),
                )

                for i, arg in enumerate(node.args):
                    # This block implements *args in Python 3.5+
                    if isinstance(arg, ast.Starred):
                        self.visit(arg)
                        continue

                    self.context.add_opcodes(
                        JavaOpcodes.DUP(),
                        ICONST_val(i),
                    )
                    self.visit(arg)
                    self.context.add_opcodes(
                        JavaOpcodes.AASTORE(),
                    )

                # This block implements *args in Python 3.4
                if getattr(node, 'starargs', None) is not None:
                    # Evaluate the starargs
                    self.visit(node.starargs)

                    self.context.add_opcodes(
                        AddToArgs(),
                    )

            if len(node.keywords) == 0 and getattr(node, 'kwargs', None) is None:
                self.context.add_opcodes(
                    JavaOpcodes.ACONST_NULL(),
                )
            else:
                # Create and populate the map of kwargs to pass to invoke().
                self.context.add_opcodes(
                    java.Map(),
                )

                for keyword in node.keywords:
                    if keyword.arg is None:  # Python 3.5 **kwargs
                        self.add_doublestarred_kwargs(node, keyword.value)
                        continue

                    self.context.add_opcodes(
                        JavaOpcodes.DUP(),
                        JavaOpcodes.LDC_W(keyword.arg),
                    )
                    self.visit(keyword.value)
                    self.context.add_opcodes(
                        java.Map.put()
                    )

                if getattr(node, 'kwargs', None) is not None:  # Python 3.4 **kwargs
                    self.add_doublestarred_kwargs(node, node.kwargs)

            # Set up the stack and invoke the callable
            self.context.add_opcodes(
                python.Callable.invoke(),
            )

    @node_visitor
    def visit_Starred(self, node):
        # Handle *args at a call site (Python 3.5+)
        # Evaluate the starargs
        self.visit(node.value)

        self.context.add_opcodes(
            AddToArgs(),
        )

    def add_doublestarred_kwargs(self, node, kwargs):
        self.visit(kwargs)

        # Add all the kwargs to the kwargs dict.
        try:
            func_name = node.func.id
        except AttributeError:
            func_name = node.func.attr

        self.context.add_opcodes(
            AddToKwargs(func_name)
        )

    @node_visitor
    def visit_Num(self, node):
        if isinstance(node.n, int):
            self.context.add_int(node.n)
        elif isinstance(node.n, float):
            self.context.add_float(node.n)
        elif isinstance(node.n, complex):
            self.context.add_complex(node.n)
        else:
            raise NotImplementedError('Unknown number type %s' % type(node.n))

    @node_visitor
    def visit_Str(self, node):
        self.context.add_str(node.s)

    @node_visitor
    def visit_Bytes(self, node):
        self.context.add_opcodes(
            java.New('org/python/types/Bytes'),

            JavaOpcodes.BIPUSH(len(node.s)),
            JavaOpcodes.NEWARRAY(JavaOpcodes.NEWARRAY.T_BYTE),
        )

        for i, b in enumerate(node.s):
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
                ICONST_val(i),
                JavaOpcodes.BIPUSH(node.s[i]),
                JavaOpcodes.BASTORE(),
            )

        self.context.add_opcodes(
            java.Init('org/python/types/Bytes', '[B')
        )

    @node_visitor
    def visit_NameConstant(self, node):
        if node.value is None:
            self.context.add_opcodes(
                python.NONE()
            )
        elif node.value is True:
            self.context.add_opcodes(
                JavaOpcodes.GETSTATIC('org/python/types/Bool', 'TRUE', 'Lorg/python/types/Bool;'),

            )
        elif node.value is False:
            self.context.add_opcodes(
                JavaOpcodes.GETSTATIC('org/python/types/Bool', 'FALSE', 'Lorg/python/types/Bool;'),

            )
        else:
            raise NotImplementedError("Unknown named constant %s" % node.value)

    @node_visitor
    def visit_Ellipsis(self, node):
        raise NotImplementedError('No handler for Ellipsis')

    @node_visitor
    def visit_Attribute(self, node, ctx=None):
        ctx = ctx or node.ctx
        self.visit(node.value)

        if type(ctx) == ast.Load:
            self.context.add_opcodes(
                python.Object.get_attribute(node.attr),
            )
        elif type(ctx) == ast.Store:
            self.context.add_opcodes(
                JavaOpcodes.SWAP(),
                python.Object.set_attr(node.attr),
            )
        elif type(ctx) == ast.Del:
            self.context.add_opcodes(
                JavaOpcodes.LDC_W(node.attr),
            )
        else:
            raise NotImplementedError("Unknown context %s" % ctx)

    @node_visitor
    def visit_Subscript(self, node, ctx=None):
        ctx = ctx or node.ctx
        if type(ctx) == ast.Load:
            self.visit(node.value)
            self.visit(node.slice)
            self.context.add_opcodes(
                python.Object.get_item()
            )
        elif type(ctx) == ast.Store:
            self.context.add_opcodes(
                ASTORE_name('#value'),
            )
            self.visit(node.value)
            self.visit(node.slice)
            self.context.add_opcodes(
                ALOAD_name('#value'),
                python.Object.set_item(),
                free_name('#value'),
            )
        elif type(ctx) == ast.Del:
            self.visit(node.value)
            self.visit(node.slice)
        else:
            raise NotImplementedError("Unknown context %s" % node.ctx)

    @node_visitor
    def visit_Name(self, node):
        if type(node.ctx) == ast.Load:
            try:
                self.context.load_name(node.id)
            except NameError:
                self.context.add_opcodes(
                    java.New('org/python/exceptions/UnboundLocalError'),
                    JavaOpcodes.LDC_W(node.id),
                    java.Init('org/python/exceptions/UnboundLocalError', 'Ljava/lang/String;'),
                    JavaOpcodes.ATHROW()
                )
        elif type(node.ctx) == ast.Store:
            self.context.store_name(node.id)
        elif type(node.ctx) == ast.Del:
            self.context.delete_name(node.id)
        else:
            raise NotImplementedError("Unknown context %s" % node.ctx)

    @node_visitor
    def visit_List(self, node):
        if isinstance(node.ctx, ast.Load):
            self.context.add_opcodes(
                python.List()
            )

            for child in node.elts:
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                )

                self.visit(child)

                self.context.add_opcodes(
                    python.List.append()
                )

        elif isinstance(node.ctx, ast.Store):
            self.context.add_opcodes(
                python.Object.iter()
            )
            for child in node.elts:
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    python.Iterable.next()
                )
                self.visit(child)
            self.context.add_opcodes(
                JavaOpcodes.POP()
            )

    @node_visitor
    def visit_Tuple(self, node):
        if isinstance(node.ctx, ast.Load):
            self.context.add_opcodes(
                java.New('org/python/types/Tuple'),

                java.List(len(node.elts))
            )

            for child in node.elts:
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                )

                self.visit(child)

                self.context.add_opcodes(
                    java.List.add()
                )

            self.context.add_opcodes(
                java.Init('org/python/types/Tuple', 'Ljava/util/List;')
            )

        elif isinstance(node.ctx, ast.Store):
            self.context.add_opcodes(
                python.Object.iter()
            )
            for child in node.elts:
                self.context.add_opcodes(
                    JavaOpcodes.DUP(),
                    python.Iterable.next()
                )
                self.visit(child)
            self.context.add_opcodes(
                JavaOpcodes.POP(),
            )

    @node_visitor
    def visit_Slice(self, node):
        self.context.add_opcodes(
            java.New('org/python/types/Slice'),
        )
        if node.lower:
            self.visit(node.lower)
        else:
            self.context.add_opcodes(
                python.NONE()
            )

        if node.upper:
            self.visit(node.upper)
        else:
            self.context.add_opcodes(
                python.NONE()
            )

        if node.step:
            self.visit(node.step)
        else:
            self.context.add_opcodes(
                python.NONE()
            )

        self.context.add_opcodes(
            java.Init('org/python/types/Slice', 'Lorg/python/Object;', 'Lorg/python/Object;', 'Lorg/python/Object;')
        )

    @node_visitor
    def visit_ExtSlice(self, node):
        # slice* dims):
        raise NotImplementedError('No handler for ExtSlice')

    @node_visitor
    def visit_Index(self, node):
        self.visit(node.value)

    @node_visitor
    def visit_ExceptHandler(self, node):
        # expr? type, identifier? name, stmt* body):
        if isinstance(node.type, ast.Tuple):
            exception = [
                self.full_classref(exc.id, default_prefix='org.python.exceptions')
                for exc in node.type.elts
            ]
        elif node.type:
            exception = self.full_classref(node.type.id, default_prefix='org.python.exceptions')
        else:
            exception = None

        self.context.add_opcodes(
            CATCH(exception),
        )

        # Top of stack is the exception to be raised
        if exception and node.name:
            # The exception has been named. Store it as that name.
            self.context.add_opcodes(
                JavaOpcodes.DUP(),
            )
            self.context.store_name(node.name)

        # Regardless of whether the exception is named,
        # locally store it so that it can be re-raised easily.
        exc_name = '#exception-%x' % id(node)
        self.context.add_opcodes(
            ASTORE_name(exc_name),
        )

        self.current_exc_name.append(exc_name)

        for child in node.body:
            self.visit(child)

        self.context.add_opcodes(
            free_name(exc_name)
        )
        self.current_exc_name.pop()



from ..utils import TranspileTestCase


class ListComprehensionTests(TranspileTestCase):
    def test_syntax(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print([v**2 for v in x])

            print([v for v in x])
            """)

    def test_list_comprehension_with_if_condition(self):
        self.assertCodeExecution("""
            print([v for v in range(100) if v % 2 == 0])
            print([v for v in range(100) if v % 2 == 0 if v % 3 == 0])
            print([v for v in range(100) if v % 2 == 0 if v % 3 == 0 if v > 10 if v < 80])
            """)

    def test_method(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(list(v**2 for v in x))
            """)



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase

from unittest import expectedFailure


class DictTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = {}
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = {}
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_creation(self):
        # Empty dict
        self.assertCodeExecution("""
            x = {}
            print(x)
            """)

        self.assertCodeExecution("""
            x = {'a': 1}
            print(x)
            """)

    def test_getitem(self):
        # Simple existent key
        self.assertCodeExecution("""
            y = 37
            x = {'a': 1, 'b': 2, 'c': y}
            print('a' in x)
            print('a' not in x)
            print(x['a'])
            """)

        # Simple non-existent key
        self.assertCodeExecution("""
            x = {'a': 1, 'b': 2}
            print('c' in x)
            print('c' not in x)
            try:
                print(x['c'])
            except KeyError as err:
                print(err)
            """)

    def test_clear(self):
        # Clear a dictionary
        self.assertCodeExecution("""
            x = {'a': 1, 'b': 2}
            print('a' in x)
            print(x.clear())
            print('a' not in x)
            print(x)
            """)

        # Clear an already empty dict
        self.assertCodeExecution("""
            x = {}
            print('a' not in x)
            print(x.clear())
            print('a' not in x)
            print(x)
            """)

    def test_builtin_constructor(self):
        # Construct a dictionary using the dict builtin
        self.assertCodeExecution("""
            x = dict()
            print(x)
            print('a' in x)

            # List of tuples
            x = dict([('a', 1), ('b', 2)])
            print('a' in x)
            print(x['a'])
            print('c' in x)

            # List of lists
            x = dict([['a', 3], ['b', 4]])
            print('a' in x)
            print(x['a'])
            print('c' in x)

            # Tuple of lists
            x = dict((['a', 5], ['b', 6]))
            print('a' in x)
            print(x['a'])
            print('c' in x)

            # Tuple of tuples
            x = dict((('a', 5), ('b', 6)))
            print('a' in x)
            print(x['a'])
            print('c' in x)

            # Test __contains__ throws unhashable exception
            try:
                print([] in x)
            except TypeError as err:
                print(err)
            try:
                print([] not in x)
            except TypeError as err:
                print(err)
        """)

    def test_builtin_constructor_kwargs(self):
        self.assertCodeExecution("""
            d = dict(a=1, b=2)
            print('a' in d)
            print('b' in d)
            print('c' not in d)
            print(d['b'])

            d = dict(d, b=3)
            print('a' in d)
            print('b' in d)
            print('c' in d)
            print(d['b'])
        """)

    def test_builtin_non_2_tuples(self):
        # One of the elements isn't a 2-tuple
        self.assertCodeExecution("""
            try:
                x = dict([('a', 1), ('b', 2, False)])
            except ValueError as err:
                print(err)
            """)

    def test_builtin_non_sequence(self):
        # One of the elements isn't a sequence
        self.assertCodeExecution("""
            try:
                x = dict([('a', 1), False, ('b', 2)])
            except TypeError as err:
                print(err)
            """)

    def test_method_pop(self):
        self.assertCodeExecution("""
            x = {1: 2, 3: 4, 5: 6}
            print(x.pop(1))
            print(x.pop(3, 37))
            try:
                print(x.pop(7))
            except KeyError as e:
                print("Dict doesn't contain 7")
            print(x.pop(7, 42))
            print("Done")
            """)

    def test_method_popitem(self):
        self.assertCodeExecution("""
            ITEMS = [(1, 2), (3, ("4", 5))]
            x = dict(ITEMS)

            popped_1 = x.popitem()
            print(popped_1 in ITEMS)

            popped_2 = x.popitem()
            print(popped_2 in ITEMS and popped_2 != popped_1)

            # Check for exception
            try:
                print(x.popitem())
            except KeyError as err:
                print(err)
            """)

    def test_method_setdefault(self):
        self.assertCodeExecution("""
            x = {42: 'Babel'}
            print(x.setdefault(42)) # should return Babel

            print(x.setdefault(1)) # should return None
            print(x[1] == None) # should be True

            print(x.setdefault('David', 'Gilmour')) # should return 'Gilmour'

            # Check unhashable exceptions
            try:
                x.setdefault([], 42)
            except TypeError as err:
                print(err)
            """)

    def test_method_get(self):
        self.assertCodeExecution("""
            x = {1: 2}
            print(x.get(1))
            print(x.get(2))
            print(x.get(3,4))

            class MyClass:
                pass

            obj = MyClass()
            print(x.get(obj))
            """)

        # check for unhashable type errors
        self.assertCodeExecution("""
            x = {1: 2}

            try:
                print(x.get([]))
            except TypeError as err:
                print(err)

            try:
                print(x.get([], 1))
            except TypeError as err:
                print(err)

            try:
                print(x.get(list([1])))
            except TypeError as err:
                print(err)

            try:
                print(x.get(set([1, 2, 3])))
            except TypeError as err:
                print(err)

            try:
                print(x.get(iter([1, 2, 3])))
            except TypeError as err:
                print(err)

            try:
                print(x.get(slice(3)))
            except TypeError as err:
                print(err)
            """)

    def test_copy(self):
        self.assertCodeExecution("""
            x = {42: 'Babel'}
            y = x.copy()

            print(y)
            print(x == y)
            print(x is not y)
            """)

        self.assertCodeExecution("""
            x = {'42': 'Babel'}
            y = x.copy()

            print(x['42'] is y['42'])

            x['42'] = 'somevalue'
            print(x['42'] == y['42'])

            print(x == y)
            """)

    def test_fromkeys(self):
        self.assertCodeExecution("""
            keys = [1, 2]
            print(dict.fromkeys(keys))
            """)

        # non-iterable error test
        self.assertCodeExecution("""
            keys = 1
            value = 2
            try:
                print(dict.fromkeys(keys, value))
            except TypeError as err:
                print(err)
            """)

        # empty iterable and orginal dict test
        self.assertCodeExecution("""
            keys = ()
            value = 5
            dt = {}
            print(dt.fromkeys(keys, value))
            print(dt)
            """)

        # non-hashable error on key test
        self.assertCodeExecution("""
            keys = [[1], 2]
            try:
                print(dict.fromkeys(keys))
            except TypeError as err:
                print(err)
            """)

    def test_update(self):
        self.assertCodeExecution("""
            a = {}
            a.update([('a', 1), ('b', 2)])
            print(sorted(a))
            b = {}
            b.update({'a': 1, 'b':2})
            print(sorted(b))
            c = {}
            c.update(a=1, b=2)
            print(sorted(c))
        """)

        self.assertCodeExecution("""
            try:
                a = {}
                a.update([('a', 1, 2), ('b',2)])
                print('An error should have been raised!')
            except ValueError:
                print('Received a ValueError as expected')
        """)

        self.assertCodeExecution("""
            try:
                a = {}
                a.update('1')
                print('An error should have been raised')
            except ValueError:
                print('Received a ValueError as expected')
        """)

        self.assertCodeExecution("""
            try:
                a = {}
                a.update(1)
                print('An error should have been raised')
            except TypeError:
                print('Received a TypeError as expected')
        """)

        self.assertCodeExecution("""
            try:
                a = {}
                x = set([1, 2])
                a.update(x)
                print('An error should have been raised')
            except TypeError:
                print('Received a TypeError as expected')
        """)

    @expectedFailure
    def test_fromkeys_missing_iterable(self):
        self.assertCodeExecution("""
            try:
                print(dict.fromkeys())
            except TypeError as err:
                print(err)
            """)

    def test_values(self):
        self.assertCodeExecution("""
            x = {1 : 1,2 : 2,3 : 3}
            y = x.values()
            print(type(y))
            print(len(y))
            print(3 in y)
            print(42 not in y)
            if y:
                print('Working as expected')
            s = 0
            for i in y:
                s += i
            print(s)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2}
                print(x.values() * 'a')
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                print(y[0])
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                print(y['a'])
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                y[0] = 7
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                del y[0]
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                print(+y)
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                print(-y)
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            try:
                x = {1 : 1,2 : 2,3 : 3}
                y = x.values()
                print(~y)
                print('A TypeError should have been raised')
            except TypeError as err:
                print(err)
            """)

    def test_items(self):
        self.assertCodeExecution("""
            x = {1 : 1,2 : 2,3 : 3}
            x2 = {1 : 1,2 : 2,3 : 3,4 : 4}
            y = x.items()
            y2 = x2.items()
            print(len(y))
            print((1,1) in y)
            print((1,2) not in y)
            s1 = 0
            s2 = 0
            for i,j in y:
                s1 += i
                s2 += j
            print(s1)
            print(s2)
            s = {(1,1),(1,2),(2,2),(2,3)}
            print(len(y & s))
            print(len(y | s))
            print(len(y ^ s))
            print(len(y - s))
            """)


class UnaryDictOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'dict'


class BinaryDictOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'dict'

    not_implemented = [
        'test_subscr_bytearray',
        'test_subscr_class',
        'test_subscr_complex',
    ]


class InplaceDictOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'dict'



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class SuperTests(TranspileTestCase):
    pass


class BuiltinSuperFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["super"]

    not_implemented = [
        'test_bool',
        'test_bytearray',
        'test_bytes',
        'test_class',
        'test_complex',
        'test_dict',
        'test_float',
        'test_frozenset',
        'test_int',
        'test_list',
        'test_None',
        'test_NotImplemented',
        'test_range',
        'test_set',
        'test_slice',
        'test_str',
        'test_tuple',
        'test_obj',
    ]



from unittest import expectedFailure

from ..utils import TranspileTestCase


class LambdaTests(TranspileTestCase):
    def test_lambda(self):
        self.assertCodeExecution("""
            f = lambda : 42
            print(f())

            g = lambda x : x
            print(g(10))

            print((lambda x: x + '!!!')('hello'))

            print((lambda x='something': x + '!!!')('hello'))

            print((lambda x: x + '!!!')(x='hello'))

            """)

    def test_lambda_kwonly(self):
        self.assertCodeExecution("""
            print((lambda *,x='something': x + '!!!')())
            """)


class FunctionTests(TranspileTestCase):
    def test_function(self):
        self.assertCodeExecution("""
            def myfunc(value):
                print(value * 3)
                return value + 5

            print("value =", myfunc(5))
            """)

    def test_function_name_attribute(self):
        self.assertCodeExecution("""
            def myfunc():
                def otherfunc():
                    pass
                print(otherfunc.__name__)

            print(myfunc.__name__)
            myfunc()
            """)

    def test_noargs_function(self):
        self.assertCodeExecution("""
            def myfunc():
                print('Hello')
                return 5

            print("value =", myfunc())
            """)

    def test_void_function(self):
        self.assertCodeExecution("""
            def myfunc(value):
                print(value * 3)

            myfunc(5)
            """)

    def test_mainline(self):
        self.assertCodeExecution("""
            if __name__ == '__main__':
                print("Hello, world")
            """, run_in_function=False)

    def test_inner_function(self):
        self.assertCodeExecution("""
            def myfunc(value):
                print(value * 3)

                def myinner(value2):
                    print(value2 * 4)
                    return value2 + 6

                print("inner value =", myinner(10))
                return value + 5

            print("outer =", myfunc(5))
            """, run_in_function=False)

    def test_closure(self):
        self.assertCodeExecution("""
            def myfunc(value):
                print(value * 3)

                def myinner(value2):
                    print(value2 * value)
                    return value2 + 6

                print("inner value =", myinner(10))
                return value + 5

            print("outer =", myfunc(5))
            """)

    def test_default_args(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(37)
            """, run_in_function=False)

    def test_override_some_default_args(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(37, 42)
            """, run_in_function=False)

    def test_overide_all_default_args(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(37, 42, 99)
            """, run_in_function=False)

    def test_use_kwargs(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(37, y=42)
            """, run_in_function=False)

    def test_use_kwargs_non_sequential(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(37, z=42)
            """, run_in_function=False)

    def test_use_all_kwargs(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(x=37, y=42, z=99)
            """, run_in_function=False)

    def test_use_all_kwargs_different_order(self):
        self.assertCodeExecution("""
            def myfunc(x, y=1, z=2):
                print("x =", x)
                print("y =", y)
                print("z =", z)

            myfunc(z=99, y=42, x=37)
            """, run_in_function=False)

    def test_call_function_with_var_args(self):
        self.assertCodeExecution("""
            def myfunc(*args):
                print(args)
                return args[0]

            print("first arg =", myfunc(1, 2, 3, 4, 5))
            """, run_in_function=False)

    def test_call_function_with_kw(self):
        self.assertCodeExecution("""
            def myfunc(**kwargs):
                print(kwargs['first'] * 3)
                print(kwargs['second'] * 3)

                return kwargs['first'] + kwargs['second']

            print("values sum =", myfunc(first=1, second=2))
            """, run_in_function=False)

    def test_call_function_kw(self):
        self.assertCodeExecution("""
            def myfunc(**kwargs):
                print(kwargs['first'] * 3)
                print(kwargs['second'] * 3)

                return kwargs['first'] + kwargs['second']

            values = {'first': 1, 'second': 2}
            print("values sum =", myfunc(**values))
            """, run_in_function=False)

    def test_call_function_var_kw(self):
        self.assertCodeExecution("""
            def myfunc(*args, **kwargs):
                print(kwargs['first'] * 3)
                print(kwargs['second'] * 3)
                print(args)

                return kwargs['first'] + kwargs['second']

            values_tuple = (1, 2, 3, 4)
            values_dict = {'first': 1, 'second': 2}
            print("values sum =", myfunc(*values_tuple, **values_dict))
            """, run_in_function=False)

    def test_call_function_var(self):
        self.assertCodeExecution("""
            def myfunc(*args):
                print(args)

                return len(args)

            values_tuple = (1, 2, 3, 4)
            print("values count =", myfunc(*values_tuple))
            """, run_in_function=False)

    def test_redefine(self):
        self.assertCodeExecution("""
            def myfunc(value):
                print(value * 3)
                return value + 5

            def myfunc(value):
                print(value * 4)
                return value + 6

            print("value =", myfunc(5))
            """)

    def test_redefine_nested_from_other_function(self):
        self.assertCodeExecution("""
            def func():
                def nested_func():
                    print("Hello World from func")

                nested_func()

            def func2():
                def nested_func():
                    print("Hello World from func2")

                nested_func()

            func()
            func2()
            """)

    def test_define_nested_generator(self):
        self.assertCodeExecution("""
            def wrapper():
                def func():
                    def gen():
                        yield 'Hello World'

                    print(next(gen()))

                func()

            wrapper()
            """)

    def test_noarg_unexpected_extra_arg(self):
        self.assertCodeExecution("""
            def myfunc():
                pass

            myfunc(5)
            """, exits_early=True)

    def test_noarg_unexpected_extra_several_args(self):
        self.assertCodeExecution("""
            def myfunc():
                return 16

            print(myfunc(1, 2, 3))
            """, exits_early=True)

    def test_unexpected_extra_several_args(self):
        self.assertCodeExecution("""
            def myfunc(a):
                return 16

            print(myfunc(1, 2, 3))
            """, exits_early=True)

    def test_multiple_args_unexpected_extra_several_args(self):
        self.assertCodeExecution("""
            def myfunc(a, b):
                return 16

            print(myfunc(1, 2, 3))
            """, exits_early=True)

    def test_missing_sole_arg(self):
        self.assertCodeExecution("""
            def myfunc(a):
                return 16

            print(myfunc())
            """, exits_early=True)

    def test_missing_args_more_args(self):
        self.assertCodeExecution("""
            def otherfunc(a, b, c, d, e, f=3):
                return 'ok'

            print(otherfunc(1))
            """, exits_early=True)

    def test_missing_args_with_varargs(self):
        self.assertCodeExecution("""
            def myfunc(a, b, *c, d=3):
                return 'ok'

            print(myfunc(1))
            """, exits_early=True)

    def test_a_method_missing_args(self):
        self.assertCodeExecution("""
            class MyClass:
                def a_method(self, a, b, c):
                    return 16

            i = MyClass()
            print(i.a_method())
            """, exits_early=True)

    def test_function_kwonly(self):
        self.assertCodeExecution("""
            def myfunc(*,x=100):
                return x

            print (myfunc())
            """)

    def test_function_try_except_return_no_exception(self):
        self.assertCodeExecution("""
            def myfunc(x):
                try:
                    return x
                except NameError:
                    return None

            print(myfunc(10))
            """)

    def test_function_try_except_return_caught_exception(self):
        self.assertCodeExecution("""
            def myfunc(x):
                try:
                    return y
                except NameError:
                    return x

            print(myfunc(10))
            """)

    def test_function_try_except_return_uncaught_exception(self):
        self.assertCodeExecution("""
            def myfunc(x):
                try:
                    return z
                except IndexError:
                    return x

            print(myfunc(10))
            """, exits_early=True)

    def test_function_frozenset_constant(self):
        self.assertCodeExecution("""
            def func():
                for i in {1, 2, 3, 4, 5, 6}:
                    print(i)

                print('a' in {1, 'a', False, 1.1, b'1', (2)})

            func()
        """)

    @expectedFailure
    def test_function_has_code_attr(self):
        self.assertCodeExecution("""
            def func():
                return 1

            if func.__code__ is not None:
                print("Code object exists")
        """)

    @expectedFailure
    def test_function_code_attr(self):
        self.assertCodeExecution("""
            def func():
                return 1

            print("__code__:", func.__code__)
            print("Function name:", func.__code__.co_name)
            print("Arg count:", func.__code__.co_argcount)
        """)




from unittest import expectedFailure

from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase
from .. utils import adjust


class SubscrSliceTest(TranspileTestCase):
    def test_subscr_str_extended(self):
        str_samples = [
                        '""',
                        '"3"',
                        '"This is another string"',
                        '"Mÿ hôvèrçràft îß fûłl öf éêlś"',
                        '"One arg: %s"',
                        '"Three args: %s | %s | %s"',
                    ]
        slice_samples = [
                        'slice(0)',
                        'slice(5)',
                        'slice(-1)',
                        'slice(-7)',
                        'slice(2, 7)',
                        'slice(-1, 2)',
                        'slice(2, -7)',
                        'slice(-2, -7)',
                        'slice(0, 0)',
                        'slice(-10, 0)',
                        'slice(0, -10)',
                        'slice(2, 7, 2)',
                        'slice(7, 2, -1)',
                        'slice(7, 2, 10)',
                        'slice(2, 7, 0)'
                    ]

        data = []
        for x in str_samples:
            for y in slice_samples:
                data.append((x, y))

        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> x = %(x)s')
                        print('>>> y = %(y)s')
                        print('>>> x[y]')
                        x = %(x)s
                        y = %(y)s
                        print(x[y])
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'x': x,
                        'y': y
                    }
                )
                for x, y in data
            ),
            "Error running test_subscr_str_extended",
            run_in_function=False,
        )


class StrTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = "Hello, world"
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_iscase(self):
        self.assertCodeExecution("""
            for s in ['hello, World!', 'HELLO, WORLD.', 'ello? world', '']:
                print(s.islower())
                print(s.isupper())
            """)

    def test_isdigit(self):
        self.assertCodeExecution("""
            for s in ['112358132134', '3.14159', '12312344df', '']:
                print(s.isdigit())
            """)

    def test_isspace(self):
        self.assertCodeExecution(r"""
            for s in ['\x1f \v \f \n \t \r', ' ', '\x85\xa0', '\u2007', '\u202f',
            '\u180e', '\t\tnope\t\t', '']:
                print(s.isspace())
            """)

    def test_isalnum(self):
        self.assertCodeExecution("""
            for word in ["", "12", "abc", "abc12", "\u00c4", "\x41", "a@g", "äÆ",
            "12.2", "'Hi'", "Hello!!", "HELLO", "V0c", "A A"]:
                print(word.isalnum())
            """)

    def test_isalpha(self):
        self.assertCodeExecution("""
            for s in ['Hello World', 'hello wORLd.', 'Hello world.', '', 'hello1',
            'this', 'this is string example....wow!!!', 'átomo', 'CasesLikeTheseWithoutSpaces']:
                print(s.isalpha())
            """)

    def test_isdecimal(self):
        self.assertCodeExecution("""
            for word in ["", "12", "abc", "abc12", "\u0037", "\x31", "0101b",
            "-13", "12.2", "'7'"]:
                print(word.isdecimal())
            """)

    def test_istitle(self):
        self.assertCodeExecution(r"""
            print("".istitle())
            print("abcd".istitle())
            print("NOT".istitle())
            print("coca cola".istitle())
            print("they are from UK, are they not?".istitle())
            print("/@.".istitle())
            print("\u00c4".istitle())
            print("\x41".istitle())
            print("py.bee".istitle())
        """)

    def test_join(self):
        self.assertCodeExecution("""
            print(','.join(None))
            """, exits_early=True)
        self.assertCodeExecution("""
            print(','.join(12))
            """, exits_early=True)
        self.assertCodeExecution("""
            print(','.join(['1', '2', '3']))
            print(','.join([]))
            print('asdf'.join(','))
            """)

    def test_endswith(self):
        self.assertCodeExecution("""
            s = "abracadabra"
            suffix = "abra"
            print(s.endswith(suffix))
            """)

        self.assertCodeExecution("""
            s = "abracadabra"
            suffix = "ABRA"
            print(s.endswith(suffix))
            """)

        self.assertCodeExecution("""
            s = "ABRACADABRA"
            suffix = "abra"
            print(s.endswith(suffix))
            """)

        self.assertCodeExecution("""
            print('abracadabra'.endswith('abra',0,5))
            """)

        self.assertCodeExecution("""
            s = "ABRACADABRA"
            suffix = ""
            print(s.endswith(suffix,3))
            """)

    def test_startswith(self):
        self.assertCodeExecution("""
            s = "abracadabra"
            suffix = "abra"
            print(s.startswith(suffix))
            """)

        self.assertCodeExecution("""
            s = "abracadabra"
            suffix = "ABRA"
            print(s.startswith(suffix))
            """)

        self.assertCodeExecution("""
            s = "ABRACADABRA"
            suffix = "abra"
            print(s.startswith(suffix))
            """)

        self.assertCodeExecution("""
            print('abracadabra'.startswith('abra',0,5))
            """)

        self.assertCodeExecution("""
            s = "ABRACADABRA"
            suffix = ""
            print(s.startswith(suffix,3))
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = "Hello, world"
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_getitem(self):
        # Simple positive index
        self.assertCodeExecution("""
            x = "12345"
            print(x[2])
            """)

        # Simple negative index
        self.assertCodeExecution("""
            x = "12345"
            print(x[-2])
            """)

        # Positive index out of range
        self.assertCodeExecution("""
            x = "12345"
            try:
                print(x[10])
            except IndexError as err:
                print(err)
            """)

        # Negative index out of range
        self.assertCodeExecution("""
            x = "12345"
            try:
                print(x[-10])
            except IndexError as err:
                print(err)
            """)

    def test_setitem(self):
        # Strings are immutable and do not allow item assignment
        self.assertCodeExecution("""
            x = "BeeWare"
            try:
                x[0] = "A"
            except TypeError as err:
                print(err)
        """)

    def test_slice(self):
        # Full slice
        self.assertCodeExecution("""
            x = "12345"
            print(x[:])
            """)

        # Left bound slice
        self.assertCodeExecution("""
            x = "12345"
            print(x[1:])
            """)

        # Right bound slice
        self.assertCodeExecution("""
            x = "12345"
            print(x[:4])
            """)

        # Slice bound in both directions
        self.assertCodeExecution("""
            x = "12345"
            print(x[1:4])
            """)

        # Slice bound in both directions with end out of bounds
        self.assertCodeExecution("""
            x = "12345"
            print(x[1:6])
            """)

        # Slice bound in both directions with start out of bounds
        self.assertCodeExecution("""
            x = "12345"
            print(x[6:7])
            """)

        # Slice bound in both directions with start larger than end
        self.assertCodeExecution("""
            x = "12345"
            print(x[4:1])
            """)

        # Slice bound in both directions with start and end out of bounds
        self.assertCodeExecution("""
            x = "12345"
            print(x[-10:10])
            """)

    def test_case_changes(self):
        self.assertCodeExecution("""
            for s in ['hello, world', 'HEllo, WORLD', 'átomo', '']:
                print(s.capitalize())
                print(s.lower())
                print(s.swapcase())
                print(s.title())
                print(s.upper())
            """)

    def test_split(self):
        self.assertCodeExecution(r"""
            for s in ['\vhello,        world', '\nHEllo, WORLD\f', 'átomo', '']:
                print(s.split())
                print(s.split("o"))
                print(s.split(maxsplit=2))
                print(s.split("l",maxsplit=0))
                try:
                    print(s.split(5))
                except TypeError as err:
                    print(err)

            """)

    def test_index(self):
        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.index('hell'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            try:
                print(s.index('world'))
            except ValueError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.index('hell', 1))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            try:
                print(s.index('hell', 1, 3))
            except ValueError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.index('hell', 1, 100))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            try:
                print(s.index('hell', 1, -1))
            except ValueError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.index('hell', -4))
            """)

    def test_count(self):
        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('e'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('a'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('ll'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('ll', 3))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('ll', 3, 4))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('ll', 0, 4))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('ll', 0, 100))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('hell', 1, -1))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.count('hell', -4))
            """)

    def test_find(self):
        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('world'))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell', 1))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell', 1, 3))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell', 1, 100))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell', 1, -1))
            """)

        self.assertCodeExecution("""
            s = 'hello hell'
            print(s.find('hell', -4))
            """)

    def test_expand(self):
        self.assertCodeExecution("""
            print('\\t'.expandtabs())
            print('a\\t'.expandtabs())
            print('aa\\t'.expandtabs())
            print('aaa\\t'.expandtabs())
            print('aaaaaaaa\\t'.expandtabs())
            print('a\\naa\\t'.expandtabs())
            print('\\t'.expandtabs(3))
            print('a\\t'.expandtabs(3))
            print('aa\\t'.expandtabs(7))
            print('aaa\\t'.expandtabs(4))
            print('aaaaaaaa\\t'.expandtabs(4))
            print('a\\naa\\t'.expandtabs(4))
            """)

    def test_title(self):
        self.assertCodeExecution(r"""
            print("".title())
            print("abcd".title())
            print("NOT".title())
            print("coca cola".title())
            print("they are from UK, are they not?".title())
            print("/@.".title())
            print("\u00c4".title())
            print("\x41".title())
            print("py.bee".title())
        """)

    def test_len(self):
        self.assertCodeExecution("""
            s = ' foo  bar    baz '
            print(len(s))
        """)

    def test_center(self):
        # test ok
        self.assertCodeExecution("""
            s = "abc"
            print(s.center(10, "-"))
            print(s.center(10))
            print(s.center(2,"-"))
            print(s.center(True,"-"))
            try:
                print(s.center(10,"-*"))
            except TypeError as err:
                print(err)
            s = 123
            try:
                print(s.center(10,"-"))
            except AttributeError as err:
                print(err)
            """)

    def test_ljust(self):
        self.assertCodeExecution("""
            s = "abc"
            print(s.ljust(5, "P"))
            print(s.ljust(2))
            print(s.ljust(15, "0"))
            print(s.ljust(20, "ã"))
            try:
                print(s.ljust(5, b'_'))
            except TypeError as err:
                print(TypeError)
            try:
                print(s.ljust(5.0, "P"))
            except TypeError as err:
                print(err)
                """)

    def test_partition(self):
        self.assertCodeExecution("""
            s = "foobar"
            print(s.partition("ob"))
            print(s.partition("o"))
            print(s.partition("f"))
            print(s.partition("r"))
            print(s.partition("x"))
            try:
                print(s.partition(""))
            except ValueError as err:
                print(err)
            """)

    def test_lstrip(self):
        self.assertCodeExecution("""
            str = "\t\t   gggfoo "
            print(str.lstrip('h'))
            print(str.lstrip())
            print(str.lstrip(None))
            print(str.lstrip(''))
            print(str.lstrip('\t '))
            print(str.lstrip('\t gf'))
            print(str.lstrip('\t\t   gggfoo '))
            try:
                print(str.lstrip(6))
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            str="abbaccdcbbs"
            print(str.lstrip('ab'))
            str=""
            print(str.lstrip())
            print(str.lstrip('ab'))
            """)

        self.assertCodeExecution(r"""
            str="\u180eabc"
            print(str.lstrip())
            str="\x85abc"
            print(str.lstrip())
            """)

    def test_rstrip(self):
        self.assertCodeExecution("""
            str = " fooggg\t\t   "
            print(str.rstrip('h'))
            print(str.rstrip())
            print(str.rstrip(None))
            print(str.rstrip(''))
            print(str.rstrip('\t '))
            print(str.rstrip('\t og'))
            print(str.rstrip(' fooggg\t\t   '))
            try:
                print(str.rstrip(6))
            except TypeError as err:
                print(err)
            """)

        self.assertCodeExecution("""
            str="abbaccdcbbsabba"
            print(str.rstrip('ab'))
            str=""
            print(str.rstrip())
            print(str.rstrip('ab'))
            """)

        self.assertCodeExecution(r"""
            str="abc\u180e"
            print(str.rstrip())
            str="abc\x85"
            print(str.rstrip())
            """)

    def test_rfind(self):
        # test cases to generate outout
        self.assertCodeExecution("""
            st="a good cook could cook good"
            print(st.rfind('cook'))
            print(st.rfind('cook',10))
            print(st.rfind('book',1,10))
            """)

        # test cases with indices more than the string length
        self.assertCodeExecution("""
            st="a good cook could cook good"
            print(st.rfind('cook',100,200))
            print(st.rfind('cook',1000))
            print(st.rfind('cook',-1))
            """)

        # test cases with empty find string
        self.assertCodeExecution("""
            st="a good cook could cook good"
            try:
                print(st.rfind())
            except TypeError as err:
                print(err)
            """)

    def test_rindex(self):
        # test cases to generate outout
        self.assertCodeExecution("""
            st="a good cook could cook good"
            try:
                print(st.rindex('cook'))
            except ValueError as err:
                print(err)
            try:
                print(st.rindex('cook',10))
            except ValueError as err:
                print(err)
            try:
                print(st.rindex('cook',1,10))
            except ValueError as err:
                print(err)
            """)

        # test cases with indices more than the string length
        self.assertCodeExecution("""
            st="a good cook could cook good"
            try:
                print(st.rindex('cook',100,200))
            except ValueError as err:
                print(err)
            try:
                print(st.rindex('cook',1000))
            except ValueError as err:
                print(err)
            try:
                print(st.rindex('cook',-1))
            except ValueError as err:
                print(err)
            """)

        # test cases with empty find string
        self.assertCodeExecution("""
            st="a good cook could cook good"
            try:
                print(st.rindex())
            except TypeError as err:
                print(err)
            """)

    def test_rjust(self):
        # test cases to generate valid outout
        self.assertCodeExecution("""
            st="a good cook could cook good"
            print(st.rjust(100))
            print(st.rjust(100, 'X'))
            print(st.rjust(10, 'X'))

            """)

        # test cases to generate exception
        self.assertCodeExecution("""
            st="a good cook could cook good"
            try:
                print(st.rjust(-20, 'X'))
            except TypeError as err:
                print(err)
            try:
                print(st.rjust(100, 'cook'))
            except TypeError as err:
                print(err)
            try:
                print(st.rjust())
            except TypeError as err:
                print(err)
            try:
                print(st.rjust(20.5, 'X'))
            except TypeError as err:
                print(err)
            """)

    def test_strip(self):
        self.assertCodeExecution("""
            s = "\t\t   hello "
            try:
                print(s.strip(6))
            except TypeError as e:
                print(e)
            print(s.strip())
            print(s.strip(None))
            print(s.strip(''))
            print(s.strip('a'))
            print(s.strip(' '))
            print(s.strip('\t '))
            print(s.strip('\t ho'))
            print(s.strip('\t hello '))
            """)

        self.assertCodeExecution("""
            str="abbaccdcbbsabba"
            print(str.strip('abs'))
            print(str.strip())
            str=""
            print(str.strip())
            print(str.strip('ab'))
            """)

    def test_casefold(self):
        self.assertCodeExecution("""
            print("ÅAÆΣß".casefold())
            print("ß.nfG".casefold())
            print("HeLlo_worldʃ!".casefold())
            """)

    def test_replace(self):
        self.assertCodeExecution("""
            s="abc abc abc abc abc abcd mm       "
            print(s.replace('','k'))
            print(s.replace(' ','k'))
            print(s.replace(' ','fm'))
            print(s.replace('ab','bc'))
            try:
                print(s.replace(45,'bc'))
            except TypeError as err:
                print(err)
            try:
                print(s.replace('kk',45))
            except TypeError as err:
                print(err)
            """)

    def test_rpartition(self):
        self.assertCodeExecution("""
            st = "Hello World!"
            print(st.rpartition("H"))
            print(st.rpartition(" "))
            print(st.rpartition("l"))
            print(st.rpartition("q"))
            print(st.rpartition("lo"))
            print(st.rpartition("!"))
            print(st.rpartition("ld!"))
            print("voc".rpartition("foobar"))
            print("".rpartition("oi"))
            class str2:
                def __str__(self):
                    return "abc"
            try:
                print(st.rpartition(str2()))
            except TypeError as err:
                print(err)
            try:
                print(st.rpartition(""))
            except ValueError as err:
                print(err)
            try:
                print(st.rpartition(4))
            except TypeError as err:
                print(err)
            try:
                print(st.rpartition(7.8))
            except TypeError as err:
                print(err)
            try:
                print(st.rpartition())
            except TypeError as err:
                print(err)
            """)

    def test_rsplit(self):
        self.assertCodeExecution(r"""
           for s in ['This is for rsplit',' eeed d ded  eded   ','átomo','',' t e s t i n g   r s p l i t ','a b c']:
                print(s.rsplit())
                print(s.rsplit("e"))
                print(s.rsplit(maxsplit=2))
                print(s.rsplit(maxsplit=-5))
                print(s.rsplit("e",-9))
                print(s.rsplit(" ", -10))
                try:
                    print(s.split(2))
                except TypeError as err:
                    print(err)
            """)

    def test_isnumeric(self):
        self.assertCodeExecution("""
        for str_ in ['123', '123.4', 'abc', '', ' ', 'ABCD', 'ABCD ', '12323445',
        '123.', '.12', '1A', 'A1', '!@#', 'A1@#']:
            print(str_.isnumeric())
            """)

    def test_isidentifier(self):
        self.assertCodeExecution("""
        for str_ in ['_sjkd', 'abc', 'ABC', 'b13s', 'foo_bar', 'eÃⱣỉ', 'ÃⱣỉ', "22222", " ", "", "/",
        "4a2a", "*", "ab cd", "!z", "&a", "@", "%"]:
            print(str_, str_.isidentifier())
            """)

    def test_isprintable(self):
        self.assertCodeExecution("""
        for str_ in [chr(i) for i in range(33)] + ['AAA', 'bcd', '1234', 'eÃⱣỉ', 'ÃⱣỉ', '', '\x07' + 'foo', '\u2029']:
            print(str_.isprintable())
            """)

    def test_isprintable_missing_cases(self):
        self.assertCodeExecution(r"""
        tests = ['\u2028', '\u2029', '\u00A0']
        for test in tests:
            print(test.isprintable())
        """)

    def test_isprintable_surrogate_cases(self):
        self.assertCodeExecution(r"""
        tests = ['\ud801', '\udbff', '\udc00', '\udfff']
        for test in tests:
            print(test.isprintable())
        """)

    def test_repr(self):
        self.assertCodeExecution(r"""
        tests = ["\r\n", "áéíóú", "\u000B", "\u2029", "\\", "'", "\"", "\"'"]
        for test in tests:
            print(repr(test))
        """)

    def test_splitlines(self):
        self.assertCodeExecution(r"""
        str_ = "aaa\nbbb\rccc\r\nddd\n\reee"

        print(str_.splitlines())
        print(str_.splitlines(True))
        print("Don't Panic\n".splitlines())
        print('\n'.splitlines())
        print(''.splitlines())

        s1 = '\r\n\r\n\v\f\x0b\x0c\u2029\x1c\x1d\x1e\x85'
        print(s1.splitlines())
        print(s1.splitlines(True))
        """)

    def test_zfill(self):
        self.assertCodeExecution("""
            s = '42'
            print(s.zfill(5))

            try:
                print(s.zfill('string'))
            except TypeError as err:
                print(err)

            try:
                print(s.zfill({}))
            except TypeError as err:
                print(err)

            s = '-42'
            print(s.zfill(5))

            s = '+42'
            print(s.zfill(5))

            s = ''
            print(s.zfill(5))

            s = '-.-42'
            print(s.zfill(6))

            s = "42"
            try:
                s.zfill()
            except TypeError as err:
                print(err)
        """)

    def test_contains(self):
        self.assertCodeExecution("""
            print('abc' in 'abc')
            print('a' in 'abc')
            print('a' in '')
            print('' in 'a')

            print('abc' not in 'abc')
            print('a' not in 'abc')
            print('a' not in '')
            print('' in 'a')
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(str(1, 2, 3, 4, 5))
            except TypeError as err:
                print(err)

            try:
                print(str("1", "2", "3", "4", "5"))
            except TypeError as err:
                print(err)

        """)


class UnaryStrOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'str'


class BinaryStrOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'str'

    not_implemented = [
        'test_modulo_class',
    ]

    is_flakey = [
        'test_modulo_obj'
    ]


class InplaceStrOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'str'

    not_implemented = [
        'test_modulo_class',
    ]

    is_flakey = [
        'test_modulo_obj'
    ]



from .. utils import TranspileTestCase, BuiltinTwoargFunctionTestCase


class DelattrTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            class MyClass(object):
                class_value = 42

                def __init__(self, val):
                    self.value = val

            print("On class: ")
            delattr(MyClass, 'class_value')
            try:
                print('  class_value =', MyClass.class_value)
                print("  Shouldn't be able to get attribute class_value")
            except AttributeError:
                print("  Can't get attribute class_value")

            try:
                delattr(MyClass, 'class_value')
                print("  Shouldn't be able to delete attribute class_value")
            except AttributeError:
                print("  Can't delete attribute class_value")

            try:
                delattr(MyClass, 'other_class_value')
                print("  Shouldn't be able to delete attribute other_class_value")
            except AttributeError:
                print("  Can't delete attribute other_class_value")

            obj = MyClass(37)

            print("On instance:")
            delattr(obj, 'value')
            try:
                print('  value =', MyClass.value)
                print("  Shouldn't be able to get attribute value")
            except AttributeError:
                print("  Can't get attribute value")

            try:
                delattr(obj, 'value')
                print("  Shouldn't be able to delete attribute value")
            except AttributeError:
                print("  Can't delete attribute value")

            try:
                delattr(obj, 'other_value')
                print("  Shouldn't be able to delete attribute other_value")
            except AttributeError:
                print("  Can't delete attribute other_value")
            """, run_in_function=False)


class BuiltinDelattrFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["delattr"]

    not_implemented = [
        'test_class_str',
        'test_obj_str',
        'test_class_obj'
    ]



from ...java import opcodes as JavaOpcodes

from .primitives import ICONST_val


##########################################################################
# Common Java operations
##########################################################################

class New:
    def __init__(self, classname):
        self.classname = classname

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.NEW(self.classname),
            JavaOpcodes.DUP()
        )


class Init:
    def __init__(self, classname, *args):
        self.classname = classname
        self.args = args

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.INVOKESPECIAL(
                self.classname,
                '<init>',
                args=self.args,
                returns='V'
            ),
        )


class Yield:
    def __init__(self, yield_point):
        self.yield_point = yield_point

    def process(self, context):
        context.add_opcodes(
            ICONST_val(self.yield_point),
            JavaOpcodes.INVOKEVIRTUAL(
                'org/python/types/Generator',
                'yield',
                args=['Ljava/util/Map;', 'I'],
                returns='V'
            ),
            # "yield" by returning from the generator method.
            JavaOpcodes.ARETURN()
        )


##########################################################################
# Java types and their operations
##########################################################################

class Array:
    def __init__(self, size, classname='org/python/Object', fill=None):
        self.size = size
        self.classname = classname
        self.fill = fill

    def process(self, context):
        context.add_opcodes(
            ICONST_val(self.size),
            JavaOpcodes.ANEWARRAY(self.classname),
        )
        if self.fill:
            for i in range(self.size):
                context.add_opcodes(
                    JavaOpcodes.DUP(),
                    ICONST_val(i),
                    self.fill,
                    JavaOpcodes.AASTORE(),
                )

    class get:
        def __init__(self, index):
            self.index = index

        def process(self, context):
            context.add_opcodes(
                ICONST_val(self.index),
                JavaOpcodes.AALOAD(),
            )


class List:
    def __init__(self, size=None):
        self.size = size

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.NEW('java/util/ArrayList'),
            JavaOpcodes.DUP(),
        )

        if self.size:
            context.add_opcodes(
                ICONST_val(self.size),
                Init('java/util/ArrayList', 'I')
            )
        else:
            context.add_opcodes(
                Init('java/util/ArrayList')
            )

    class add:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/List',
                    'add',
                    args=['Ljava/lang/Object;'],
                    returns='Z'
                ),
                JavaOpcodes.POP(),
            )

    class addAll:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/List',
                    'addAll',
                    args=['Ljava/util/Collection;'],
                    returns='Z'
                ),
                JavaOpcodes.POP(),
            )


class Map:
    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.NEW('java/util/HashMap'),
            JavaOpcodes.DUP(),
            Init('java/util/HashMap')
        )

    class get:
        def __init__(self, key):
            self.key = key

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.key),
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/Map',
                    'get',
                    args=['Ljava/lang/Object;'],
                    returns='Ljava/lang/Object;'
                )
            )

    class put:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/Map',
                    'put',
                    args=['Ljava/lang/Object;', 'Ljava/lang/Object;'],
                    returns='Ljava/lang/Object;'
                ),
                JavaOpcodes.POP()
            )

    class putAll:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/Map',
                    'putAll',
                    args=['Ljava/util/Map;'],
                    returns='V'
                ),
            )

    class remove:
        def __init__(self, key):
            self.key = key

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.key),
                JavaOpcodes.INVOKEINTERFACE(
                    'java/util/Map',
                    'remove',
                    args=['Ljava/lang/Object;'],
                    returns='Ljava/lang/Object;'
                ),
                JavaOpcodes.POP()
            )


class Class:
    class forName:
        def __init__(self, classname):
            self.classname = classname

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.classname),
                JavaOpcodes.INVOKESTATIC(
                    'java/lang/Class',
                    'forName',
                    args=['Ljava/lang/String;'],
                    returns='Ljava/lang/Class;'
                ),
            )


class THROW:
    # Raise an exception of given type with given arguments
    # Example:
    #    THROW(
    #        'org/python/exceptions/AttributeError',
    #        ['Ljava/lang/String;', JavaOpcodes.LDC_W("Invalid attribute")],
    #    )
    def __init__(self, exception_class, *exception_args):
        self.exception_class = exception_class
        self.exc_arg_types = [e[0] for e in exception_args]
        self.exc_arg_values = [e[1] for e in exception_args]

    def process(self, context):
        context.add_opcodes(
            New(self.exception_class),
            *self.exc_arg_values
        )
        context.add_opcodes(
            Init(self.exception_class, *self.exc_arg_types),
            JavaOpcodes.ATHROW(),
        )



#!/usr/bin/env python

"""Microbenchmark for function call overhead.

This measures simple function calls that are not methods, do not use varargs or
kwargs, and do not use tuple unpacking.
"""

# Python imports
import time


def foo(a, b, c, d):
    # 20 calls
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)
    bar(a, b, c)


def bar(a, b, c):
    # 20 calls
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)
    baz(a, b)


def baz(a, b):
    # 20 calls
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)
    quux(a)


def quux(a):
    # 20 calls
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()
    qux()


def qux():
    pass


def test_calls(iterations):
    times = []
    for _ in range(iterations):
        t0 = time.time()
        # 40 calls
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        foo(1, 2, 3, 4)
        t1 = time.time()
        times.append(t1 - t0)
    return times


if __name__ == "__main__":
    # Priming run.
    test_calls(1)

    import sys
    times = test_calls(int(sys.argv[1]))

    print("Time elapsed: " + str(sum(times)))



from .. utils import TranspileTestCase, BuiltinFunctionTestCase

from unittest import expectedFailure

class LocalsTests(TranspileTestCase):
    def test_simple(self):
        self.assertCodeExecution("""
            print("There are %s locals" % len(locals()))
            x = 1
            y = 'z'
            print("There are %s locals" % len(locals()))

            def method():
                print("In method: there are %s locals" % len(locals()))

                try:
                    print("locals()['x'] =", locals()['x'])
                except KeyError:
                    print("Variable x not defined")
                try:
                    print("locals()['y'] =", locals()['y'])
                except KeyError:
                    print("Variable y not defined")
                try:
                    print("locals()['z'] =", locals()['z'])
                except KeyError:
                    print("Variable z not defined")

                x = 1
                y = 'z'
                locals()[y] = 2

                print("locals()['x'] =", locals()['x'])
                print("locals()['y'] =", locals()['y'])
                print("locals()['z'] =", locals()['z'])
                print('x', x)
                print('y', y)
                # print('z', z)

                print("In method: there are %s locals" % len(locals()))

            method()

            print("There are %s locals" % len(locals()))
            print("locals()['x'] =", locals()['x'])
            print("locals()['y'] =", locals()['y'])
            try:
                print("locals()['z'] =", locals()['z'])
            except KeyError:
                print("Variable z not defined")
            print('x', x)
            print('y', y)

            print('Done')
        """, run_in_function=False)

    @expectedFailure
    def test_preloaded_locals(self):
        self.assertCodeExecution("""
            lcls = locals
            x = 1
            y = 'z'

            print("There are %s locals" % len(lcls()))

            def method():
                print("In method: there are %s locals" % len(lcls()))

                x = 1
                y = 'z'

                print("In method: there are %s locals" % len(lcls()))

            method()
        """, run_in_function=False)


class BuiltinLocalsFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["locals"]

    not_implemented = [
        'test_bool',
        'test_bytearray',
        'test_bytes',
        'test_class',
        'test_complex',
        'test_dict',
        'test_float',
        'test_frozenset',
        'test_int',
        'test_list',
        'test_None',
        'test_NotImplemented',
        'test_range',
        'test_set',
        'test_slice',
        'test_str',
        'test_tuple',
        'test_obj',
    ]



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase

from unittest import expectedFailure


class IntTests(TranspileTestCase):
    def test_int(self):
        self.assertCodeExecution("""
            for i in range(-500, 500):
                print(i)
            """)

    def test_setattr(self):
        self.assertCodeExecution("""
            x = 37
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = 37
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = 37
            try:
                x[0] = 1
            except TypeError as err:
                print(err)
        """)

    def test_invalid_literal(self):
        self.assertCodeExecution("""
            try:
                print(int('q', 16))
            except ValueError as err:
                print(err)
            """)

    def test_none(self):
        self.assertCodeExecution("""
            try:
                print(int(None))
            except TypeError as err:
                print(err)
            """)

    def test_base_correct(self):
        self.assertCodeExecution("""
            print(int('a', 16))
            print(int('22', 3))
            print(int('5', 0))
            """)

    def test_base_none(self):
        self.assertCodeExecution("""
            try:
                print(int('0', None))
            except TypeError as err:
                print(err)
            """)

    def test_base_too_small(self):
        self.assertCodeExecution("""
            try:
                print(int('22', 1))
            except ValueError as err:
                print(err)
            """)

    def test_base_negative(self):
        self.assertCodeExecution("""
            try:
                print(int('5', -10))
            except ValueError as err:
                print(err)
            """)

    def test_base_too_large(self):
        self.assertCodeExecution("""
            try:
                print(int('5', 100))
            except ValueError as err:
                print(err)
            """)

    def test_no_arguments(self):
        self.assertCodeExecution("""
            print(int())
            """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(int('1', 2, 3, 4))
            except TypeError as err:
                print(err)
            try:
                print(int('1', 2, 3, 4, 5))
            except TypeError as err:
                print(err)
            """)


class UnaryIntOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'int'


class BinaryIntOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'int'

    not_implemented = [
    ]


class InplaceIntOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'int'

    substitutions = {
        # Float tests rounding
        "(6.134323640217668+11.638718805964096j)": [
            "(6.134323640217669+11.638718805964096j)"
        ],
        "(20.329893907303543+38.57212830945331j)": [
            "(20.329893907303546+38.57212830945331j)"
        ],
        # Complex tests rounding
        "(0.7037573231697811+0.7104404479510611j)": [
            "(0.703757323169781+0.7104404479510611j)"
        ],
        "(-1.758764802874057+2.4303798814529753j)": [
            "(-1.7587648028740568+2.430379881452976j)"
        ],
        "(-0.9523946730227619-0.3048678185671067j)": [
            "(-0.9523946730227619-0.30486781856710676j)"
        ],
        "(4669868.983337471+4714215.686526724j)": [
            "(4669868.98333747+4714215.686526724j)"
        ],
        "(-4.926088756782315e-05-1.5768735125131076e-05j)": [
            "(-4.926088756782315e-05-1.576873512513108e-05j)"
        ],
        "(-35416596.83160829+5518228.872766545j)": [
            "(-35416596.83160829+5518228.872766544j)"
        ],
    }



from .. utils import TranspileTestCase, BuiltinTwoargFunctionTestCase


class FilterTests(TranspileTestCase):

    def test_bool(self):
        self.assertCodeExecution('print(list(filter(bool, [True, False, True])))')
        self.assertCodeExecution('print(list(filter(bool, [1, 0, 3, -1])))')
        self.assertCodeExecution('print(list(filter(bool, [])))')

    def test_none(self):
        self.assertCodeExecution('print(list(filter(None, [True, False, True])))')
        self.assertCodeExecution('print(list(filter(None, [])))')

    def test_lambda(self):
        self.assertCodeExecution('print(list(filter(lambda x: x > 1, [3, 4, 56, 1, -11])))')

    def test_wrong_argument(self):
        self.assertCodeExecution('print(list(filter(None, None)))', exits_early=True)


class BuiltinFilterFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["filter"]

    not_implemented = [
        'test_bool_bytearray',
        'test_bool_bytes',
        'test_bool_dict',
        'test_bool_frozenset',
        'test_bool_list',
        'test_bool_range',
        'test_bool_set',
        'test_bool_str',
        'test_bool_tuple',

        'test_bytearray_bytearray',
        'test_bytearray_bytes',
        'test_bytearray_dict',
        'test_bytearray_frozenset',
        'test_bytearray_list',
        'test_bytearray_range',
        'test_bytearray_set',
        'test_bytearray_str',
        'test_bytearray_tuple',

        'test_bytes_bytearray',
        'test_bytes_bytes',
        'test_bytes_dict',
        'test_bytes_frozenset',
        'test_bytes_list',
        'test_bytes_range',
        'test_bytes_set',
        'test_bytes_str',
        'test_bytes_tuple',

        'test_complex_bytearray',
        'test_complex_bytes',
        'test_complex_dict',
        'test_complex_frozenset',
        'test_complex_list',
        'test_complex_range',
        'test_complex_set',
        'test_complex_str',
        'test_complex_tuple',

        'test_dict_bytearray',
        'test_dict_bytes',
        'test_dict_dict',
        'test_dict_frozenset',
        'test_dict_list',
        'test_dict_range',
        'test_dict_set',
        'test_dict_str',
        'test_dict_tuple',

        'test_float_bytearray',
        'test_float_bytes',
        'test_float_dict',
        'test_float_frozenset',
        'test_float_list',
        'test_float_range',
        'test_float_set',
        'test_float_str',
        'test_float_tuple',

        'test_frozenset_bytearray',
        'test_frozenset_bytes',
        'test_frozenset_dict',
        'test_frozenset_frozenset',
        'test_frozenset_list',
        'test_frozenset_range',
        'test_frozenset_set',
        'test_frozenset_str',
        'test_frozenset_tuple',

        'test_int_bytearray',
        'test_int_bytes',
        'test_int_dict',
        'test_int_frozenset',
        'test_int_list',
        'test_int_range',
        'test_int_set',
        'test_int_str',
        'test_int_tuple',

        'test_list_bytearray',
        'test_list_bytes',
        'test_list_dict',
        'test_list_frozenset',
        'test_list_list',
        'test_list_range',
        'test_list_set',
        'test_list_str',
        'test_list_tuple',

        'test_NotImplemented_bytearray',
        'test_NotImplemented_bytes',
        'test_NotImplemented_dict',
        'test_NotImplemented_frozenset',
        'test_NotImplemented_list',
        'test_NotImplemented_range',
        'test_NotImplemented_set',
        'test_NotImplemented_str',
        'test_NotImplemented_tuple',

        'test_range_bytearray',
        'test_range_bytes',
        'test_range_dict',
        'test_range_frozenset',
        'test_range_list',
        'test_range_range',
        'test_range_set',
        'test_range_str',
        'test_range_tuple',

        'test_set_bytearray',
        'test_set_bytes',
        'test_set_dict',
        'test_set_frozenset',
        'test_set_list',
        'test_set_range',
        'test_set_set',
        'test_set_str',
        'test_set_tuple',

        'test_slice_bytearray',
        'test_slice_bytes',
        'test_slice_dict',
        'test_slice_frozenset',
        'test_slice_list',
        'test_slice_range',
        'test_slice_set',
        'test_slice_str',
        'test_slice_tuple',

        'test_str_bytearray',
        'test_str_bytes',
        'test_str_dict',
        'test_str_frozenset',
        'test_str_list',
        'test_str_range',
        'test_str_set',
        'test_str_str',
        'test_str_tuple',

        'test_tuple_bytearray',
        'test_tuple_bytes',
        'test_tuple_dict',
        'test_tuple_frozenset',
        'test_tuple_list',
        'test_tuple_range',
        'test_tuple_set',
        'test_tuple_str',
        'test_tuple_tuple',

        'test_obj_tuple',
        'test_obj_str',
        'test_obj_set',
        'test_obj_range',
        'test_obj_list',
        'test_obj_frozenset',
        'test_obj_dict',
        'test_obj_bytearray',
        'test_obj_bytes',

        'test_class_obj',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class NextTests(TranspileTestCase):
    def test_next_success(self):
        self.assertCodeExecution("""
            i = iter([1])
            print(next(i))
        """)

    def test_next_success_with_default(self):
        self.assertCodeExecution("""
            i = iter([1])
            print(next(i, 0))
        """)

    def test_next_exhausted_with_default(self):
        self.assertCodeExecution("""
            i = iter([])
            print(next(i, 0))
        """)

    def test_next_exhausted_without_default(self):
        self.assertCodeExecution("""
            i = iter([])
            try:
                print(next(i))
            except StopIteration as err:
                print(err)
            print('Done.')
        """)


class BuiltinNextFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["next"]

    not_implemented = [
    ]



import voc
import argparse

from .transpiler import transpile


def main():
    parser = argparse.ArgumentParser(
        prog='voc',
        description='Transpiles Python code to Java class files.'
    )

    parser.add_argument(
        '-o', '--output',
        help='The directory where class files should be output.',
        default='.'
    )
    parser.add_argument(
        '-p', '--prefix',
        help='The prefix to strip from all source file paths.',
        default='.'
    )
    parser.add_argument(
        '-n', '--namespace',
        help='The namespace for the generated Java classfiles.',
        default='python'
    )
    parser.add_argument(
        '-v', '--verbosity',
        action='count',
        default=0
    )
    parser.add_argument(
        '--version',
        action='version',
        version='voc %s' % voc.__version__,
    )
    parser.add_argument(
        'input',
        metavar='source file',
        nargs='+',
        help='The source file or directory to compile'
    )

    args = parser.parse_args()

    transpile(
        input=args.input,
        prefix=args.prefix,
        outdir=args.output,
        namespace=args.namespace,
        verbosity=args.verbosity
    )


if __name__ == '__main__':
    main()
